
/* ============================================================ *
 * lensing.h							*
 *								*
 * Martin Kilbinger, Karim Benabed 2006-2009			*
 * ============================================================ */


#ifndef __LENSING_H
#define __LENSING_H

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "errorlist.h"
#include "maths.h"
#include "io.h"
#include "cosmo.h"
#include "nofz.h"


/* Dimensions of interpolation tables */
#define N_s     200
#define N_theta 100

/* Ranges of interpolation tables */
#define s_min     1.0e-2
#define s_max     1.0e6


#define lensing_base                     -1400
#define lensing_inconsistent             -1 + lensing_base
#define lensing_baryon_fraction          -2 + lensing_base
#define lensing_tomoij			 -3 + lensing_base
#define lensing_initialised              -4 + lensing_base
#define lensing_unknown                  -5 + lensing_base
#define lensing_pm                       -6 + lensing_base

/* If Ob/Oc > BARYON_FRAC, chi2 produces an error */
#define BARYON_FRAC 0.75

#define Nlensdata_t 8
typedef enum {xipm, xip, xim, xiE, map2poly, map2gauss, gsqr, nofz} lensdata_t;
#define slensdata_t(i) ( \
 i==xipm      ? "xipm" : \
 i==xip       ? "xip"  : \
 i==xim       ? "xim"  : \
 i==xiE       ? "xiE"  : \
 i==map2poly  ? "map2poly" : \
 i==map2gauss ? "map2gauss" : \
 i==gsqr      ? "gsqr" : \
 i==nofz      ? "nofz" : \
 "")

typedef enum {benjamin, jan} lensformat_t;
#define slensformat_t(i) ( \
 i==benjamin ? "benjamin" : \
 i==jan      ? "jan" : \
 "")
#define Nlensformat_t 2


typedef enum {tomo_all, tomo_auto_only, tomo_cross_only} tomo_t;
#define stomo_t(i) ( \
 i==tomo_all        ? "tomo_all" : \
 i==tomo_auto_only  ? "tomo_auto_only" : \
 i==tomo_cross_only ? "tomo_cross_only" : \
 "")
#define Ntomo_t 3

typedef struct {

  /* Basic cosmology */
  cosmo *cosmo;

  /* Redshift distribution(s) */
  redshift_t *redshift;

  tomo_t tomo;
  char stomo[64];

  /* ============================================================ *
   * Precomputed stuff.						  *
   * ============================================================ */

  interTable **g_source;
  interTable **Pkappa;

  /* Shear second-order functions */
  interTable **xiP, **xiM, **gamma, **map_gauss, **map_poly;
} cosmo_lens;

typedef struct {
  double r;
  cosmo_lens* self;
} cosmo_lensANDdouble;

typedef struct {
  int i;
  double r;
  cosmo_lens *self;
} cosmo_lensANDintANDdouble;

typedef struct {
  int i, j;
  double r;
  cosmo_lens *self;
} cosmo_lensANDiid;

typedef struct {
  int i, j, t;
  double r;
  cosmo_lens *self;
} cosmo_lensANDiiid;

typedef struct {
  int N;	     /* total nb of data points, e.g. for xipm: xip and *
		      * xim have N/2 data points each                   */
  double *theta;     /* bin centers in arcmin */
} bin_data;

typedef struct {
  int Ntheta, Nzbin;  /* Number of angular and redshift bins */
  int Nzcorr;         /* Number of z-correlations, Nzcorr=Nzbin*(Nzbin+1)/2 */
  int n;              /* Number of total entries in data vector, n=Ntheta*Nzcorr */
  double *theta;      /* n-dimensional vector of angular scales */
  double *data;       /* n-dimensional data vector */
  double *var;        /* n-dimensional vector with variance */
  double *cov;        /* nxn-dimensional covariance matrix */
  int usecov;
  lensdata_t type;
} datcov;

/* ============================================================ *
 * Initialisation.						*
 * ============================================================ */

cosmo_lens *init_parameters_lens(double OMEGAM, double OMEGAV, double W0_DE, double W1_DE, 
				 double H100, double OMEGAB, double OMEGANUMASS,
				 double NEFFNUMASS, double NORM, double NSPEC,
				 int Nzbin, const int *Nnz, const nofz_t *nofz, double *par_nz,
				 nonlinear_t NONLINEAR, transfer_t TRANSFER,
				 growth_t GROWTH, de_param_t DEPARAM,
				 norm_t normmode, tomo_t TOMO, error **err);
cosmo_lens* copy_parameters_lens_only(cosmo_lens* source, error **err);
cosmo_lens* copy_parameters_lens(cosmo_lens* source, sm2_error **err);
void updateFrom_lens(cosmo_lens* avant, cosmo_lens* apres, error **err);
void read_cosmological_parameters_lens(cosmo_lens **self, FILE *F, error **err);
cosmo_lens* set_cosmological_parameters_to_default_lens(error **err);
void free_parameters_lens(cosmo_lens** self);
void dump_param_lens(cosmo_lens* self, FILE *F, error **err);

/* ============================================================ *
 * Lensing functions.						*
 * ============================================================ */

/* Projection */
double int_for_g(double aprime, void *intpar, error **err);
double g_source(cosmo_lens*, double a, int n_bin, error **err);
double G(cosmo_lens* self, double a, int n_bin, error **err);
double int_for_p_2(double a, void *intpar,error **err);
double Pkappa(cosmo_lens *self, double a, int i_bin, int j_bin, error **err);
double P_projected_kappa(void *self, double l, int i_bin, int j_bin, error **err);

/* Second-order shear functions */
double xi(cosmo_lens*, int pm, double theta, int i_bin, int j_bin, error **err);
double gamma2(cosmo_lens*, double theta, int i_bin, int j_bin, error **err);
double map2_poly(cosmo_lens*, double theta, int i_bin, int j_bin, error **err);
double map2_gauss(cosmo_lens*, double theta, int i_bin, int j_bin, error **err);
double int_for_map2_slow(double ell, void *intpar, error **err);
double map2_slow(cosmo_lens *self, double theta, tpstat_t tpstat, int i_bin, int j_bin, error **err);

/* Index functions for tomography */
int idx_zz(int i_bin, int j_bin, int Nzbin);
int idx_tzz(int i, int i_bin, int j_bin, int Ntheta, int Nzbin);
int idx_tzztzz(int i, int i_bin, int j_bin, int j, int k_bin, int l_bin, int Ntheta, int Nzbin);

/* Reading data files */
datcov* init_data_cov(char* dataname, char* covname, lensdata_t type, int isnofz, error **err);
datcov* init_data_cov_tomo(char* dataname, char* covname, lensdata_t type, lensformat_t format, error **err);
void del_data_cov(datcov** dc);
void read_data(datcov *dc, char data_name[], error **err, int isnofz);
void read_data_tomo(datcov *dc, char data_name[], int Nzbin, error **err);
void read_cov_tomo_benjamin(datcov* dc, char cov_name[], error **err);
void read_cov_tomo_jan(datcov* dc, char cov_name[], error **err);
void read_cov(datcov* dc, char cov_name[], error **err);
void read_cov_col(datcov *dc, char cov_name[], error **err);

double lensing_signal(cosmo_lens *model, double theta, int i_bin, int j_bin, lensdata_t type, error **err);
double chi2_lensing(cosmo_lens* csm, datcov* dc, error **err);

#define CHANGE(fct) int change_##fct(cosmo_lens*, cosmo_lens*)
CHANGE(g_source);
CHANGE(Pkappa);
CHANGE(xi);
CHANGE(gamma2);
CHANGE(map2);
#undef CHANGE


#endif /* __LENSING_H */

