/* ============================================================ *
 * cmb_bao.c							*
 * Martin Kilbinger 2008					*
 * Contains the CMB 'distance priors' (Komatsu et al.2008, K08) *
 * and BAO distance ratios (e.g. Eisenstein et al. 2005, E05,   *
 * Percival et al. 2007, P07).					*
 * ============================================================ */

#include "cmb_bao.h"

#define __FUNC__ "int_for_r_sound"
double int_for_r_sound(double a, void *intpar, error **err)
{
   cosmo *self;
   double E, res, h2;

   if (a==0) return 0.0;

   self = (cosmo*)intpar;
   E    = sqrt(Esqr(self, a, 1, err));
   forwardError(*err, __LINE__, -1.0);

   h2  = self->h_100*self->h_100;
   res = a*a*E*sqrt(1.0 + a*0.75*self->Omega_b/(omega_gamma/h2));
   //res = a*a*E*sqrt(1.0 + a*0.75*self->Omega_b*3.0e4);

   if (res<=0) return 0.0;

   res = res;
   return 1.0/res;

}
#undef __FUNC__

/* Comoving sound horizon [Mpc/h], K08, Per05 */
#define __FUNC__ "r_sound"
double r_sound(cosmo *self, double a, error **err)
{
   double rs;

   rs = sm2_qromberg(int_for_r_sound, (void*)self, 0.0, a, 1.0e-7, err);
   forwardError(*err, __LINE__, -1.0);

   return rs*R_HUBBLE/sqrt(3.0);
}
#undef __FUNC__

/* ============================================================ *
 * Hu & Sugiyama (1996) fitting formula for the decoupling red- *
 * shift, see also K08 (66-68).					*
 * ============================================================ */
double z_star(cosmo *self)
{
   double g1, g2, ob, om, h2, zs;

   h2 = self->h_100*self->h_100;
   ob = self->Omega_b*h2;
   om = self->Omega_m*h2;

   g1 = 0.0783*pow(ob, -0.238)/(1.0 + 39.5*pow(ob, 0.763));
   g2 = 0.560*(1.0 + 21.1*pow(ob, 1.81));
   zs = 1048.0*(1.0 + 0.00124*pow(ob, -0.738))*(1.0 + g1*pow(om, g2));

   return zs;
}

/* Drag epoch, Eisenstein & Hu (1998) */
double z_drag(cosmo *self)
{
   double zd, om, ob, h2, b1, b2;

   h2 = self->h_100*self->h_100;
   ob = self->Omega_b*h2;
   om = self->Omega_m*h2;

   b1 = 0.313*pow(om, -0.419)*(1.0 + 0.607*pow(om, 0.674));
   b2 = 0.238*pow(om, 0.223);
   zd  = 1291.0*pow(om, 0.251)/(1.0 + 0.659*pow(om, 0.828))
     *(1.0 + b1*pow(ob, b2));

   return zd;
}


/* K08 (65) */
#define __FUNC__ "acoustic_scale"
double acoustic_scale(cosmo *self, error **err)
{
   double zs, l_A, ws, as, r_s;

   zs   = z_star(self);
   as   = 1.0/(1.0+zs);
   ws   = w(self, as, 1, err);     forwardError(*err, __LINE__, -1.0);

   /* f_K is comoving angular diameter distance, therefore no (1+z_*) prefactor */
   l_A  = pi*f_K(self, ws, err);   forwardError(*err, __LINE__, -1.0);
   r_s  = r_sound(self, as, err);  forwardError(*err, __LINE__, -1.0);
   l_A /= r_s;

   return l_A;
}
#undef __FUNC__

/* K08 (69) */
#define __FUNC__ "shift_parameter"
double shift_parameter(cosmo *self, error **err)
{
   double R, zs, ws, as;

   zs = z_star(self);
   as = 1.0/(1.0+zs);
   ws = w(self, as, 0, err);                      forwardError(*err, __LINE__, -1.0);
   R  = sqrt(self->Omega_m)*f_K(self, ws, err);   forwardError(*err, __LINE__, -1.0);
   R /= R_HUBBLE;

   return R;
}
#undef __FUNC__

/* Spherically average distance, E05, eq. (2), in units of Mpc/h */
#define __FUNC__ "D_V"
double D_V(cosmo *self, double a, error **err)
{
   double DV, ww;

   ww  = w(self, a, 0, err);                     forwardError(*err, __LINE__, -1.0);
   DV  = dsqr(f_K(self, ww, err));		 forwardError(*err, __LINE__, -1.0);
   DV /= sqrt(Esqr(self, a, 0, err));		 forwardError(*err, __LINE__, -1.0);
   DV *= R_HUBBLE*(1.0/a - 1.0); //*self->h_100;

   return cbrt(DV);
}
#undef __FUNC__

#define __FUNC__ "chi2_cmbDP"
double chi2_cmbDP(cosmo *model, mvdens *g, error **err)
{
   double *x, c2;
   size_t ndim;

   ndim = g->ndim;

   testErrorRet(ndim!=3 && ndim!=4, ce_npar,
		"Number of dimensions in likelihood for CMB distance prior has to be 3 or 4",
		*err, __LINE__, 0.0);

   x = malloc_err(sizeof(double)*ndim, err);          forwardError(*err, __LINE__, 0.0);

   //dump_param(model, stderr);

   x[0] = acoustic_scale(model, err);                 forwardError(*err, __LINE__, 0.0);
   x[1] = shift_parameter(model, err);                forwardError(*err, __LINE__, 0.0);
   x[2] = z_star(model);

   //fprintf(stderr, "x[0] = %g\n", x[0]);
   //fprintf(stderr, "x[1] = %g\n", x[1]);
   //fprintf(stderr, "x[2] = %g\n", x[2]);

   if (ndim==4) {
      /* 100*omega_b */
      x[3] = 100.0*model->Omega_b*model->h_100*model->h_100;
   }

   c2 = mvdens_log_pdf(g, x, err);
   forwardError(*err, __LINE__, 0.0);

   free(x);

   return c2;
}
#undef __FUNC__

#define N_BAO 2
#define __FUNC__ "chi2_bao"
double chi2_bao(cosmo *model, mvdens *g, error **err)
{
   /* SDSS & 2dfGRS redshifts */
   const double z_BAO[N_BAO] = {0.2, 0.35};
   double *x, c2, r_s, z_d;
   int ndim, i;

   ndim = g->ndim;

   testErrorRet(ndim!=N_BAO, ce_npar,
		"Number of dimensions in likelihood for CMB distance prior has to be 3 or 4",
		*err, __LINE__, 0.0);

   x = malloc_err(sizeof(double)*ndim, err);    forwardError(*err, __LINE__, 0.0);

   testErrorRet(ndim!=N_BAO, ce_npar,
		"Number of dimensions in likelihood for CMB distance prior has to be 3 or 4",
		*err, __LINE__, 0.0);

   z_d  = z_drag(model);
   r_s  = r_sound(model, 1.0/(z_d+1.0), err);    forwardError(*err, __LINE__, 0.0);

   /* BAO data vector: x_i = r_s/D_V(z_i) */
   for (i=0; i<N_BAO; i++) {
      x[i] = r_s/D_V(model, 1.0/(z_BAO[i]+1.0), err);  forwardError(*err, __LINE__, 0.0);
      /* Correction to match P07 sound horizon? */
      //x[i] *= 1.00129;
      //fprintf(stderr, "x[%.2f] = %.5f    ", z_BAO[i], x[i]);
   }

   c2 = mvdens_log_pdf(g, x, err);
   forwardError(*err, __LINE__, 0.0);

   free(x);

   return c2;
}
#undef __FUNC__
#undef N_BAO

#define N_BAO 1
#define __FUNC__ "chi2_bao_only"
double chi2_bao_only(cosmo *model, mvdens *g, error **err)
{
   double x, c2;
   int ndim;

   ndim = g->ndim;

   testErrorRet(ndim!=N_BAO, ce_npar,
		"Number of dimensions in likelihood for CMB distance prior has to be 3 or 4",
		*err, __LINE__, 0.0);

   /* BAO data: x = D_V(0.35)/D_V(0.2) */
   x  =  D_V(model, 1.0/(0.35+1.0), err);  forwardError(*err, __LINE__, 0.0);
   x /=  D_V(model, 1.0/(0.20+1.0), err);  forwardError(*err, __LINE__, 0.0);

   c2 = mvdens_log_pdf(g, &x, err);
   forwardError(*err, __LINE__, 0.0);

   testErrorRet(!isfinite(c2), ce_infnan, "inf or nan in logl", *err, __LINE__, 0.0);

   return c2;
}
#undef __FUNC__
#undef N_BAO

#define N_BAO 1
#define __FUNC__ "chi2_bao_A"
double chi2_bao_A(cosmo *model, mvdens *g, error **err)
{
   /* SDSS LRG redshift */
   const double z_BAO[N_BAO] = {0.35};
   double *x, c2;
   int i, ndim;

   ndim = g->ndim;

   testErrorRet(ndim!=N_BAO, ce_npar, "Wrong number of BAO redshifts",
		*err, __LINE__, 0.0);

   x = malloc_err(sizeof(double)*ndim, err);    forwardError(*err, __LINE__, 0.0);

   //dump_param(model, stderr);

   /* BAO data: x = D_V(z)/cz * sqrt(O_M) * H_0 */
   for (i=0; i<N_BAO; i++) {
      x[i]  =  D_V(model, 1.0/(z_BAO[i]+1.0), err);  forwardError(*err, __LINE__, 0.0);
      x[i] *= sqrt(model->Omega_m)/z_BAO[i]/R_HUBBLE;
   }

   c2 = mvdens_log_pdf(g, x, err);
   forwardError(*err, __LINE__, 0.0);
   free(x);

   testErrorRet(!isfinite(c2), ce_infnan, "inf or nan in logl", *err, __LINE__, 0.0);

   return c2;
}
#undef __FUNC__
#undef N_BAO

