/* ============================================================ *
 * cosmo.c							*
 * Martin Kilbinger, Karim Benabed 2006, 2007, 2008		*
 *								*
 * Early versions of the power spectrum code are based on a     *
 * program by P. Schneider.					*
 *								*
 * References:							*
 * - Press et al., Numerical Recipes in C (NR)			*
 * - Carroll, Press & Turner 1992 (CPT)				*
 * - Peacock & Dodds 1994, 1996 (PD2, PD)			*
 * - Bardeen, Bond, Kaiser & Szalay 1986 (BBKS)			*
 * - Bond & Efstathiou 1984					*
 * - Smith et al. 2003						*
 * - Eisenstein & Hu 1998					*
 * - Sugijyama 1995						*
 * - Heath et al. 1997						*
 * - Wetterich 2004						*
 * - Percival 2005						*
 * - Linder & Jenkins 2005					*
 * - Jassal, Bagla & Padmanabhan 2005				*
 * ============================================================ */


#include "cosmo.h"


/* ============================================================ *
 * Creates a new cosmo model with all tables set to 0.		*
 * ============================================================ */

#define __FUNC__ "init_parameters"
cosmo* init_parameters(double OMEGAM, double OMEGADE, double W0_DE, double W1_DE, 
		       double H100, double OMEGAB, double OMEGANUMASS,
		       double NEFFNUMASS, double NORM, double NSPEC,
		       nonlinear_t NONLINEAR, transfer_t TRANSFER, growth_t GROWTH,
		       de_param_t DEPARAM,norm_t normmode, double AMIN, 
		       error **err)
{
   cosmo* res;

   testErrorRetVA(H100<=0.0, ce_negative,
		  "Negateive Hubble parameter h_100 = %g", *err, __LINE__, NULL, H100);
   testErrorRetVA(OMEGAM<0.0, ce_negative,
		  "Negative matter density parameter Omega_m = %g", *err, __LINE__, NULL, OMEGAM);
   testErrorRetVA(OMEGAB<0.0, ce_negative,
		  "Negative baryon density parameter Omega_b = %g", *err, __LINE__, NULL, OMEGAB);
   testErrorRetVA(OMEGANUMASS<0.0, ce_negative,
		  "Negative density parameter of massive neutrinos Omega_nu_mass = %g", *err,
		  __LINE__, NULL, OMEGANUMASS);
   testErrorRetVA(NORM<0.0, ce_negative,
		  "Negative normalization = %g", *err, __LINE__, NULL, NORM);

   res = (cosmo*) malloc_err(sizeof(cosmo), err);   forwardError(*err, __LINE__, NULL);

   res->Omega_m       = OMEGAM;
   res->Omega_de      = OMEGADE;
   res->w0_de         = W0_DE;
   res->w1_de         = W1_DE;
   res->h_100         = H100;
   res->Omega_b       = OMEGAB;
   res->Omega_nu_mass = OMEGANUMASS;
   res->Neff_nu_mass  = NEFFNUMASS;
   res->normalization = NORM;
   res->n_spec        = NSPEC;

   res->nonlinear     = NONLINEAR;
   res->transfer      = TRANSFER;
   res->growth        = GROWTH;
   res->de_param      = DEPARAM;
   res->normmode      = normmode;
	 
   if (AMIN>0.0) res->a_min = AMIN;
   else res->a_min = a_minmin;

   /* Reset pre-computed numbers and tables */
   res->N_a                  = _N_a*4;
   res->linearGrowth         = NULL;
   res->growth_delta0        = -1;
   res->transferFct          = NULL;
   res->tsqr_vinschter       = NULL;
   res->transfer_alpha_Gamma = -1;
   res->transfer_s           = -1;
   res->transferBE           = NULL;
   res->cmp_sigma8           = -1;
   res->P_NL                 = NULL;
   res->slope                = NULL;
   res->w                    = NULL;
   res->k_NL                 = NULL;

   //dump_param(res, stdout);

   set_norm(res, err);
   forwardError(*err, __LINE__, NULL);

   return res;
}
#undef __FUNC__

#define __FUNC__ "copy_parameters_only"
cosmo* copy_parameters_only(cosmo* source, error **err)
{
   cosmo* res;
   res = init_parameters(source->Omega_m,source->Omega_de, source->w0_de,source->w1_de, 
			 source->h_100, source->Omega_b, source->Omega_nu_mass, source->Neff_nu_mass,
			 source->normalization, source->n_spec, source->nonlinear, source->transfer,
			 source->growth, source->de_param, source->normmode, source->a_min, err);
   forwardError(*err,__LINE__,NULL);
   return res;
}
#undef __FUNC__

#define __FUNC__ "copy_parameters"
cosmo* copy_parameters(cosmo* source, error **err)
{
   cosmo* res;

   res = init_parameters(source->Omega_m,source->Omega_de, source->w0_de,source->w1_de, source->h_100,
			 source->Omega_b, source->Omega_nu_mass, source->Neff_nu_mass,
			 source->normalization, source->n_spec, source->nonlinear,source->transfer,
			 source->growth, source->de_param, source->normmode, source->a_min, err);

   forwardError(*err,__LINE__,NULL);
   res->N_a = source->N_a;
   res->linearGrowth = copy_interTable(source->linearGrowth,err);
   forwardError(*err,__LINE__,NULL);
   res->transferFct = copy_interTable(source->transferFct,err);
   forwardError(*err,__LINE__,NULL);
   res->transferBE = copy_interTable(source->transferBE,err);
   forwardError(*err,__LINE__,NULL);
   res->tsqr_vinschter = copy_interTable2Dneq(source->tsqr_vinschter, err);
   forwardError(*err, __LINE__, NULL);
   res->P_NL = copy_interTable2D(source->P_NL,err);
   forwardError(*err,__LINE__,NULL);
   res->slope = copy_interTable(source->slope,err);
   forwardError(*err,__LINE__,NULL);
   res->w = copy_interTable(source->w,err);
   forwardError(*err,__LINE__,NULL);
   res->k_NL = copy_interTable(source->k_NL, err);
   forwardError(*err,__LINE__,NULL);
   res->growth_delta0 = source->growth_delta0;
   res->cmp_sigma8 = source->cmp_sigma8;
   res->transfer_s = source->transfer_alpha_Gamma;

   return res;
}
#undef __FUNC__

#define __FUNC__ "read_cosmological_parameters"
void read_cosmological_parameters(cosmo **self, FILE *F, error **err)
{
   cosmo *tmp;
   config_element c;

   tmp = set_cosmological_parameters_to_default(err);
   forwardError(*err, __LINE__,);

   CONFIG_READ(tmp, Omega_m, d, F, c, err);
   CONFIG_READ(tmp, Omega_de, d, F, c, err);
   CONFIG_READ(tmp, w0_de, d, F, c, err);
   CONFIG_READ(tmp, w1_de, d, F, c, err);
   CONFIG_READ(tmp, h_100, d, F, c, err);
   CONFIG_READ(tmp, Omega_b, d, F, c, err);
   CONFIG_READ(tmp, Omega_nu_mass, d, F, c, err);
   CONFIG_READ(tmp, Neff_nu_mass, d, F, c, err);
   CONFIG_READ(tmp, normalization, d, F, c, err);
   CONFIG_READ(tmp, n_spec, d, F, c, err);

   CONFIG_READ(tmp, nonlinear, i, F, c, err);
   CONFIG_READ(tmp, transfer, i, F, c, err);
   CONFIG_READ(tmp, growth, i, F, c, err);
   CONFIG_READ(tmp, de_param, i, F, c, err);
   CONFIG_READ(tmp, normmode, i, F, c, err);

   CONFIG_READ(tmp, a_min, d, F, c, err);

   *self = copy_parameters_only(tmp, err);
   forwardError(*err, __LINE__,);
}
#undef __FUNC__

#define __FUNC__ "set_cosmological_parameters_to_default"
cosmo *set_cosmological_parameters_to_default(error **err)
{
   /* Parameters are:
      Om Od w0 w1 h Ob Onu Neffnu s8 ns
      nonlin transfer growth deparam norm amin
   */

   return init_parameters(0.25, 0.75, -1.0, 0.0, 0.70, 0.044, 0.0, 0.0, 0.80, 0.96,
			  smith03, eisenhu, growth_de, linder, norm_s8, 0.0, err);
}

#define __FUNC__ "set_cosmological_parameters_to_default"
cosmo *set_cosmological_parameters_to_default2(error **err)
{
   /* Parameters are:
      Om Od w0 w1 h Ob Onu Neffnu s8 ns
      nonlin transfer growth deparam norm amin
   */

   return init_parameters(0.25, 0.75, -1.0, 0.0, 0.7, 0.04307, 0.0, 0.0, 0.80, 0.96,
			  smith03, eisenhu, growth_de, linder, norm_s8, 1.0/1211.0, err);
}

#undef __FUNC__
void free_parameters(cosmo** self)
{
   cosmo *s;

   s = *self;

   del_interTable(&s->linearGrowth);
   del_interTable(&s->transferFct);
   del_interTable(&s->transferBE);
   del_interTable2Dneq(&s->tsqr_vinschter);
   del_interTable2D(&s->P_NL);
   del_interTable(&s->slope);
   del_interTable(&s->w);
   del_interTable(&s->k_NL);

   free(s);
   s = NULL;
}

#define __FUNC__ "updateParameters"
void updateParameters(cosmo* model, double OMEGAM, double OMEGADE, double W0_DE, double W1_DE, 
		      double H100, double OMEGAB, double OMEGANUMASS,
		      double NEFFNUMASS, double NORM, double NSPEC,
		      nonlinear_t NONLINEAR, transfer_t TRANSFER, growth_t GROWTH,
		      de_param_t DEPARAM, norm_t normmode, double AMIN, error **err)
{
   cosmo* prev;

   prev = copy_parameters_only(model, err);
   forwardError(*err,__LINE__,);

   model->Omega_m   = OMEGAM;
   model->Omega_de  = OMEGADE;
   model->w0_de     = W0_DE;
   model->w1_de     = W1_DE;
   model->h_100     = H100;
   model->Omega_b   = OMEGAB;
   model->Omega_nu_mass = OMEGANUMASS;
   model->Neff_nu_mass  = NEFFNUMASS;
   model->normalization = NORM;
   model->n_spec        = NSPEC;

   model->nonlinear = NONLINEAR;
   model->transfer  = TRANSFER;
   model->growth    = GROWTH;
   model->de_param  = DEPARAM;
   model->normmode  = normmode;

   if (AMIN>0.0) model->a_min     = AMIN;
   else model->a_min = a_minmin;

   updateFrom(prev, model, err); forwardError(*err,__LINE__,);
   set_norm(model, err);         forwardError(*err,__LINE__,);
   
   free_parameters(&prev);
}
#undef __FUNC__

#define __FUNC__ "updateFrom"
void updateFrom(cosmo* avant, cosmo* apres, error **err)
{
   if (change_D_plus(avant,apres)) {
      del_interTable(&(apres->linearGrowth));
      apres->growth_delta0=-1;
   }
   if (change_Delta_L_BE2(avant,apres)) {
      del_interTable(&(apres->transferBE));
   }
   if (change_Tsqr(avant,apres)) {
      del_interTable(&(apres->transferFct));
      apres->transfer_alpha_Gamma = -1;
      apres->transfer_s = -1;
      del_interTable2Dneq(&(apres->tsqr_vinschter));
   }
   if (change_sigma_8_sqr(avant,apres)) {
      apres->cmp_sigma8=-1;
   }
   if (change_P_NL(avant,apres)) {
      del_interTable2D(&(apres->P_NL));
      del_interTable(&(apres->slope));
   }
   if (change_w(avant,apres)) {
      del_interTable(&(apres->w));
   }
   if (change_norm(avant, apres)) {
      set_norm(apres, err);
      forwardError(*err,__LINE__,);
   }
}
#undef __FUNC__

int change_norm(cosmo* avant, cosmo* apres) {
   if (NCOEQ(avant,apres,normmode) || NCOCLOSE(avant,apres,normalization))
     return 1;
   if (change_D_plus(avant,apres) || change_sigma_8_sqr(avant,apres)) return 1;
   return 0;
}

#define __FUNC__ "set_norm"
void set_norm(cosmo* self, error **err)
{
   double d0,nrm;

   // MKDEBUG
   testErrorRet(self->normmode==norm_as, ce_badFormat,
		"norm_as at the moment not supported", *err, __LINE__,);
   self->sigma_8 = self->normalization;
   self->As      = 0.0;
   return;
   
   
   if (self->growth==camb_vinschter_gr) {
     d0=1;
   }
   else {
     d0  = D_plus(self, 1.0, 0, err);             
     forwardError(*err,__LINE__,);
   }
   if ((self->transfer==camb_vinschter)&&(self->tsqr_vinschter==NULL)) {
     nrm=1;
   }
   else {
     nrm = sigma_8_sqr(self, err);
     //fprintf(stderr, "nrm=%g\n", nrm);
     forwardError(*err,__LINE__,);
   }
   if (self->normmode==norm_s8) {
     self->sigma_8 = self->normalization;
     self->As      = self->sigma_8*self->sigma_8/d0/d0/nrm;
     return;
   }
   if (self->normmode==norm_as) {
     self->As      = self->normalization;
     self->sigma_8 = sqrt(self->As * d0*d0*nrm);
     return;
   }

   *err = addError(ce_unknown, "unknown normmode", *err, __LINE__);
}
#undef __FUNC__

void dump_param(cosmo* self, FILE *F)
{
   if (!F) F = stderr;
   fprintf(F, "#  O_m    O_de   w0_de  w1_de  h_100  O_b    O_nu   Neffnu sig_8  A_s        n_s   "
	   "n t g d m\n");
   fprintf(F, "# % .3f % .3f % .3f % .3f % .3f % .3f % .3f % .3f % .3f % .3e % .3f "
	   "%d %d %d %d %d\n",
	   self->Omega_m, self->Omega_de, self->w0_de, self->w1_de, self->h_100, self->Omega_b,
	   self->Omega_nu_mass, self->Neff_nu_mass, self->sigma_8, self->As, self->n_spec,
	   self->nonlinear, self->transfer, self->growth, self->de_param, self->normmode);
}

void dump_param2(cosmo* self, FILE *F, char* pream)
{
   if (!F) F = stderr;
   fprintf(F,"%s:: \n", pream);
   dump_param(self, F);
}


/* ============================================================ *
 * Cosmology.							*
 * ============================================================ */

#define __FUNC__ "da_dtau"
/* CPT 9, no dark energy, no neutrinos */
double da_dtau(cosmo* self,double a, error **err)
{
   double res, det;

   testErrorRet(a<epsilon, ce_underflow, "Division by 0, the scale factor a is too small",
		*err, __LINE__, 0.0);
   det = 1.0 + (self->Omega_m+self->Omega_nu_mass)*(1.0/a - 1.0);
   forwardError(*err,__LINE__,0);
   det	+= self->Omega_de*(dsqr(a) - 1.0);
   forwardError(*err,__LINE__,0);
   res = sqrt(det);
   return res;
}
#undef __FUNC__

#define __FUNC__ "da_dtau_m3"
double da_dtau_m3(double a, void *intpar, error **err)
{
   double res;
   if (a < epsilon) return 0.0;
   res = da_dtau((cosmo*) intpar,a,err);
   forwardError(*err,__LINE__,0);
   testErrorRet(res<epsilon, ce_underflow, "Division by 0", *err, __LINE__, 0.0);
   return 1.0/(res*res*res);
}
#undef __FUNC__

/* Wetterich 2004, Phys. Lett. B, 594, 17 */
#define Omegadeinf0 1.0e-5
#define __FUNC__ "b_early"
double b_early(double w0, double Omegam, double Omegadeinf, error **err)
{
   double b;

   testErrorRet(Omegadeinf<=0 || Omegadeinf>=1, ce_negative, "Omegadeinf (w1_de) outside [0;1]",
		*err, __LINE__, 0.0);

   if (Omegadeinf<Omegadeinf0) {
      b = -3.0*w0/(-log(Omegadeinf) + log((1.0-Omegam)/Omegam));
   } else {
      b = -3.0*w0/(log((1.0-Omegadeinf)/Omegadeinf) + log((1.0-Omegam)/Omegam));
   }

   return b;
}
#undef __FUNC__

#define __FUNC__ "w_de"
double w_de(cosmo* self, double a, error **err)
{
   double res, b;

   res = self->w0_de;

   switch (self->de_param) {
      case jassal  :
	 if (fabs(self->w1_de)>epsilon) res += self->w1_de*a*(1.0-a);
	 break;
      case linder  :
	 if (fabs(self->w1_de)>epsilon) res += self->w1_de*(1.0-a);
	 break;
      case earlyDE :
	 /* TODO: neutrinos? */
	 b = b_early(self->w0_de, self->Omega_m, self->w1_de, err);
	 /* b = self->w1_de; */
	 forwardError(*err, __LINE__, 0.0);
	 testErrorRet(fabs(1.0-b*log(a))<epsilon, ce_singularValue, "Division by zero",
		      *err, __LINE__, 0.0);
	 res = self->w0_de/dsqr(1.0-b*log(a));
	 break;
      default     : 
	 *err = addError(ce_unknown, "Unknown de_param value", *err, __LINE__);
	 return 0;
	 break;
   }

   return res;
}
#undef __FUNC__

#define __FUNC__ "f_de"
/* Per05 (4) */
double f_de(cosmo* self,double a,error **err)
{
   double res, b;

   switch (self->de_param) {
      case jassal :
	 res = -3.0*(1.0 + self->w0_de);
	 if (fabs(self->w1_de)>epsilon)
	   res += 3.0*self->w1_de/(2.0*log(a))*(1.0-a)*(1.0-a);
	 break;
      case linder :
	 res = -3.0*(1.0 + self->w0_de);
	 if (fabs(self->w1_de)>epsilon)
	   res += 3.0*self->w1_de*((a-1.0)/log(a) - 1.0);
	 break;
      case earlyDE :
	 /* TODO: neutrinos? */
	 b   = b_early(self->w0_de, self->Omega_m, self->w1_de, err);
	 /* b = self->w1_de; */
	 forwardError(*err, __LINE__, 0.0);
	 testErrorRet(fabs(1.0+b*log(a))<epsilon, ce_singularValue, "Division by zero",
		      *err, __LINE__, 0.0);
	 res = -3.0*(1.0 + self->w0_de/(1.0-b*log(a)));
	 break;
      default     : 
	 *err = addError(ce_unknown,"Unknown de_param value",*err,__LINE__);
	 return 0;
	 break;
   }

   return res;
}
#undef __FUNC__

#define __FUNC__ "Esqr"
/* Per05 (3) */
/* MKDEBUG new: with Omega_r>0 (needed for sound horizon at z_*, z_d) 
 * Seems buggy!! */
double Esqr(cosmo* self, double a, int wOmegar, error **err)
{
   double asqr, EE, Omega_r;

   asqr = a*a;

   //fprintf(stderr, "E^2: Om = %g Ode = %g a = %g\n", self->Omega_m, self->Omega_de, a);

   if (wOmegar==1) {
      /* For early-Universe calculations like r_sound(z_*) */
      Omega_r = omega_gamma/self->h_100/self->h_100*(1.0 + 0.2271*NEFF);
   } else {
      /* Omega_r produces too much power for PD96 !!?!! */
      Omega_r = 0.0;
   }


   EE = self->Omega_m/(asqr*a)
     + (1.0 - self->Omega_m - self->Omega_de - self->Omega_nu_mass - Omega_r)/asqr
     + self->Omega_de*pow(a, f_de(self, a, err))
     + Omega_r/(asqr*asqr);
   forwardError(*err,__LINE__,0);

   if (self->Omega_nu_mass>epsilon) EE += self->Omega_nu_mass/pow(a, 3.0*(1+w_nu_mass(self, a, err)));
   forwardError(*err, __LINE__, 0);

   testErrorRetVA(EE<0, ce_negative, "E^2 = (H/H_0)^2 is negative %g %g", *err, __LINE__, 0.0, Omega_r, a);
   testErrorRetVA(!finite(EE), ce_infnan, "E^2 not finite %g %g %g %g", *err, __LINE__, 0.0,
		  self->Omega_m, (1.0 - self->Omega_m - self->Omega_de - self->Omega_nu_mass - Omega_r)/asqr,
		  self->Omega_de*pow(a, f_de(self, a, err)),  Omega_r/(asqr*asqr)
		  );

   return EE;
}
#undef __FUNC__

#define __FUNC__ "Omega_m_a"
/* Per05 (6) */
double Omega_m_a(cosmo* self,double a, double Esqrpre, error **err)
{
   double EE;
   if (Esqrpre>0) EE = Esqrpre;
   else {
      EE = Esqr(self, a, 0, err);
      forwardError(*err,__LINE__,0);
   }
   return self->Omega_m/(a*a*a)/EE;
}
#undef __FUNC__

#define __FUNC__ "Omega_de_a"
/* Per05 (6) */
double Omega_de_a(cosmo* self,double a, double Esqrpre, error **err)
{
   double EE,res;

   if (Esqrpre>0) EE = Esqrpre;
   else {
      EE = Esqr(self, a, 0, err);
      forwardError(*err, __LINE__, 0);
   }
   res = self->Omega_de*pow(a, f_de(self, a, err))/EE;
   forwardError(*err,__LINE__,0);

   return res;
}
#undef __FUNC__

#define __FUNC__ "w_nu_mass"
double w_nu_mass(cosmo *self, double a, error **err)
{
   double m_nu, w, m0, alf, bet;

   m_nu = 93.0*dsqr(self->h_100)*self->Omega_nu_mass/3.0;

   m0  = 0.000585;
   alf = 1.652;
   bet = 0.561;
   w = pow(1.0+pow(m_nu/m0*a, alf), -bet)/3.0;

   /* 1/3 ok for m_nu <= 0.05 eV */
   //return (1+0.618232*pow(m_nu*a,0.753005))/(1+9.58117*pow(m_nu*a,1.34954));
   return w;
}
#undef __FUNC__

#define __FUNC__ "D_plus_derivs"
void D_plus_derivs(double a, double *y, double *yp, void* extra, error **err)
{
   double q, r, om, ode, EE;
   cosmo* self;

   self = (cosmo*) extra;
   testErrorRet(a==0.0, ce_underflow,
		"The scale factor a has to be larger than zero. Check a_min (struct cosmo)",
		*err, __LINE__,);

   EE  = Esqr(self, a, 0, err);          forwardError(*err,__LINE__,);
   om  = Omega_m_a(self, a, EE, err);    forwardError(*err,__LINE__,);
   ode = Omega_de_a(self, a, EE, err);   forwardError(*err,__LINE__,);
   /* TODO: neutrinos */

   q = (2.0 - 0.5*(om + (1.0+3.0*w_de(self,a,err))*ode))/a;
   forwardError(*err,__LINE__,);
   r = 1.5*om/a/a;

   yp[1] = y[2];
   yp[2] = -q*y[2] + r*y[1];
}
#undef __FUNC__

int change_D_plus(cosmo* avant, cosmo* apres)
{
   if (NCOEQ(avant,apres,growth) || NCOEQ(avant,apres,N_a))
     return 1;
   switch (avant->growth) {
      case heath :
	 if (NCOCLOSE(avant,apres,Omega_m) || NCOCLOSE(avant,apres,Omega_de) ||
	     NCOCLOSE(avant,apres,a_min)) return 1;
      case growth_de :
         /* TODO: Neff_nu?? */
	 if (change_w(avant,apres)) return 1;
      case camb_vinschter_gr :
	 return change_Tsqr_camb_vinschter(avant, apres);
   }
   return 0;
}

#define nvar 2
#define eps  1.0e-8
#define __FUNC__ "D_plus"
double D_plus(cosmo* self, double a, int normalised, error **err)
{

   double delta, aa, res;
   int  i;
   double *table;
   double ystart[nvar+1], a1, a2, h1, hmin, a1min;
   int nok, nbad;
   double da;
   interTable *linearGrowth;

   if (self->linearGrowth == NULL) {
      da = (1.0 - self->a_min)/(self->N_a-1.0);
      linearGrowth = init_interTable(self->N_a,self->a_min,1.,da,0.0,0.0,err);
      forwardError(*err,__LINE__,0);
      table=linearGrowth->table;
      switch (self->growth) {
       
	 case heath :
	    self->growth_delta0 = 2.5*self->Omega_m
	      *sm2_qromberg(da_dtau_m3, (void*)self, 0.0, 1.0, 1.0e-6, err);
	    forwardError(*err,__LINE__,0);
	    for (i=0,aa=self->a_min; i<self->N_a; i++, aa+=da) {
	       delta = 2.5*self->Omega_m
		 *sm2_qromberg(da_dtau_m3, (void*)self, 0.0, aa, 1.0e-6, err);
	       forwardError(*err,__LINE__,0);
	       table[i] = da_dtau(self,aa,err)/aa*delta;
	       forwardError(*err,__LINE__,0);
	    }
	    break;
				
	 case growth_de :
	    a1min = 0.0001;
	    if (self->de_param!=earlyDE) {
	       ystart[1] = 5.0/3.0*a1min;   /* D (EdS)  */
	    } else {
	       ystart[1] = 5.0/3.0*pow(a1min, 1.0-3.0*Omega_de_a(self, a1min, -1, err)/5.0);
	       forwardError(*err, __LINE__, 0.0);
	    }
	    ystart[2] = 0.0;             /* D'(EdS)  */
	    h1   = 0.0001;
	    hmin = 0.0;

	    odeint(ystart, nvar, a1min, self->a_min, eps, h1, (void*) self, hmin, &nok, &nbad,
		   D_plus_derivs, bsstep, err);
	    forwardError(*err,__LINE__,0);
					
	    table[0] = ystart[1];

	    for (i=1,a1=self->a_min,a2=a1+da; i<self->N_a; i++,a1+=da,a2+=da) {
	       odeint(ystart, nvar, a1, a2, eps, h1, (void*) self, hmin, &nok, &nbad,
		      D_plus_derivs, bsstep, err);
	       forwardError(*err,__LINE__,0);
	       table[i] = ystart[1];
	    }
	    self->growth_delta0 = table[self->N_a-1];
	    break;

	 case camb_vinschter_gr :
	    *err = addError(ce_unknown, "D+ called with growth_t=camb_vinschter_gr",*err,__LINE__);
	    return 0.0;
	    break;

	 default : 
	    *err = addError(ce_unknown,"Unknown growth",*err,__LINE__);
	    return 0;
	    break;
      }
      self->linearGrowth = linearGrowth;
   }

   res = interpol_wr(self->linearGrowth, a, err);
   forwardError(*err,__LINE__,0);

   if (normalised==1) return res/self->growth_delta0;
   else return res;
}
#undef __FUNC__
#undef nvar
#undef eps

#define __FUNC__ "Omega_a"
/* PD 96 - (15,16) */
void Omega_a(cosmo* self, double a, double *omega_m, double *omega_v, error **err)
{
   double f, a3;
   a3 = a*a*a;
   f = a + self->Omega_m*(1.-a) + self->Omega_de*(a3-a);
   *omega_m = self->Omega_m/f;
   *omega_v = self->Omega_de*a3/f;
}
#undef __FUNC__

#define __FUNC__ "g"
/* CPT 29, PD 15+16, actually: Lahav 1991, Lightman&Schechter 1990 */
double g(cosmo* self, double a, error **err)
{
   double omega_m, omega_v;
   Omega_a(self,a, &omega_m, &omega_v,err);
   forwardError(*err,__LINE__,0);

   return 2.5*omega_m/(pow(omega_m,4.0/7.0) - omega_v +
		       (1.0 + 0.5*omega_m)*(1.0 + omega_v/70.0));
}
#undef __FUNC__


/* ============================================================ *
 * Transfer function. k in h/Mpc.				*
 * ============================================================ */
int change_Tsqr(cosmo* avant, cosmo* apres)
{
   if (NCOEQ(avant,apres,transfer))
     return 1;
   switch (apres->transfer) {
      case bbks : case eisenhu : case camb :
	 if (NCOCLOSE(avant,apres,Omega_m) || NCOCLOSE(avant,apres,h_100) || NCOCLOSE(avant,apres,Omega_b)
	     || NCOCLOSE(avant,apres,n_spec) || NCOCLOSE(avant,apres,Omega_nu_mass) || NCOCLOSE(avant,apres,Neff_nu_mass))
	   return 1;
         /* TODO: neutrinos, T_nu from BBKS */
	 return 0;
      case camb_vinschter :
	 return change_Tsqr_camb_vinschter(avant, apres);
   }
   return 0;
}

int change_Tsqr_camb_vinschter(cosmo *avant, cosmo *apres)
{
   if (NCOCLOSE(avant,apres,Omega_m) || NCOCLOSE(avant,apres,h_100) || NCOCLOSE(avant,apres,Omega_b)
       || NCOCLOSE(avant,apres,n_spec) || change_w(avant,apres) || NCOCLOSE(avant,apres,Neff_nu_mass))
     return 1;
   return 0;
}

#define Tcmb  2.725
#define __FUNC__ "Tsqr_one"
double Tsqr_one(cosmo *self, double k, double Gamma_eff,error **err)
{
   double f1, f2, q, res, L, C;

   switch (self->transfer) {

      case bbks :
 
	 q = k/Gamma_eff;
         f1 = log(1 + 2.34*q)/(2.34*q);
         f2 = 1 + q*(3.89 + q*(259.21 + q*(162.771336 + q*2027.16958081)));
         res = dsqr(f1)/sqrt(f2)*pow(k, self->n_spec);
         break;

      case eisenhu : 

	 /* [q] = [kk] = h/Mpc. EH98 (28) */
	 q = k * dsqr(Tcmb/2.7)/Gamma_eff;
	 /* W.Hu's problem set 4 */
	 /* L = log(2.71828 + 1.84*q*alpha_Gamma); */
	 /* C = 14.4 + 325./(1. + 60.5*pow(q, 1.11)); */
	 /* EH98 (29) */
	 L = log(2.*2.71828 + 1.8*q);
	 C = 14.2 + 731.0/(1.0 + 62.5*q);
	 res = dsqr(L/(L + C*q*q))*pow(k, self->n_spec);
	 break;

      case camb_vinschter : case camb :
	 *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
	 return 0;

      default :
	 *err = addError(ce_unknown,"unknown transfer",*err,__LINE__);
	 return 0; 
   }

   return res;
}
#undef __FUNC__

#define Nk 100  
#define __FUNC__ "Tsqr"
double Tsqr(cosmo* self, double k, error **err) {

   double dlogk, logkmin, logkmax;
   double *table;
   interTable *transferFct;
   double kk, f1, Gamma_eff, omhh, f_b;
   int i;

   if (self->transferFct == NULL) {
      logkmin = log(k_min);
      logkmax = log(k_max);
      dlogk   = (logkmax-logkmin)/(Nk-1.0);
      transferFct = init_interTable(N_k,logkmin,logkmax,dlogk,self->n_spec,self->n_spec-4.,err);
      forwardError(*err,__LINE__,0);
      table = transferFct->table;

      testErrorRet(self->h_100<0.0 || self->Omega_m<0.0 || self->Omega_nu_mass<0.0 || self->Omega_b<0.0,
		   ce_negative,
		   "h_100 or density parameter negative", *err, __LINE__, 0.0);
      
      switch (self->transfer) {
	 case bbks:
	    /* TODO: neutrinos */
	    Gamma_eff = self->h_100*self->Omega_m*exp(-self->Omega_b
			 - sqrt(self->h_100/0.5)*self->Omega_b/self->Omega_m);
	    for (i=0,kk=k_min; i<Nk; i++,kk*=exp(dlogk)) {
	       table[i] = log(Tsqr_one(self, kk, Gamma_eff,err));
	       forwardError(*err,__LINE__,0);
	    }
	    break;
	 case eisenhu:
	    /* TODO: neutrinos */
	    omhh = self->Omega_m*self->h_100*self->h_100;
  	    f_b  = self->Omega_b/self->Omega_m;
  	    /* sound horizon in Mpc/h, EH98 (26)*/
	    if (self->transfer_s<0)
	      self->transfer_s = 44.5*log(9.83/omhh)
		/sqrt(1.0 + 10.0*pow(self->Omega_b*dsqr(self->h_100), 0.75))*self->h_100;

  	    /* EH98 (31) */
	    if (self->transfer_alpha_Gamma<0)
	      self->transfer_alpha_Gamma = 1 - 0.328*log(431*omhh)*f_b 
		+ 0.38*log(22.3*omhh)*f_b*f_b;

	    for (i=0,kk=k_min; i<Nk; i++,kk*=exp(dlogk)) {

	       /* EH98 (30) */
	       Gamma_eff = self->Omega_m*self->h_100*(self->transfer_alpha_Gamma + 
		      (1.0-self->transfer_alpha_Gamma)/(1.0+dsqr(dsqr(0.43*kk*self->transfer_s))));
	       table[i] = log(Tsqr_one(self, kk, Gamma_eff, err));
	       forwardError(*err, __LINE__, 0);

	    }
	    break;
	 case camb_vinschter : case camb :
	    *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
	    return 0;
	 default : 
	    *err = addError(ce_unknown, "unknown transfer", *err, __LINE__);
	    return 0;
      }
      self->transferFct = transferFct;
   }

   if (k>k_max && self->transfer==bbks) {
      Gamma_eff = self->h_100*self->Omega_m*exp(-self->Omega_b - sqrt(self->h_100/0.5)
						*self->Omega_b/self->Omega_m);
      f1 = Tsqr_one(self, k, Gamma_eff, err);
      forwardError(*err,__LINE__,0);
   } else if (k>k_max && self->transfer==eisenhu) {
      testErrorRet(self->transfer_alpha_Gamma<0 || self->transfer_s<0,
		   ce_negative, "Transfer function variables not initialised", *err, __LINE__, 0.0);
      Gamma_eff = self->Omega_m*self->h_100*(self->transfer_alpha_Gamma + 
		      (1.0-self->transfer_alpha_Gamma)/(1.0+dsqr(dsqr(0.43*k*self->transfer_s))));
      f1 = Tsqr_one(self, k, Gamma_eff, err);
      forwardError(*err,__LINE__,0);
   } else {
      f1 = interpol_wr(self->transferFct,log(k),err);
      forwardError(*err,__LINE__,0);
      f1 = exp(f1);
   }
   return f1;
}
#undef Nk
#undef Tcmb
#undef __FUNC__

#define __FUNC__ "init_and_copy_Tsqr_vinschter_table"
interTable2Dspline *init_and_copy_Tsqr_vinschter_table(int m, int n, const double *x,  const double *y, const double ***z,
						       error **err)
{
   interTable2Dspline *self;
   int comp, j,im,jn;
   
   self = init_interTable2Dspline(m, n, err);
   forwardError(*err, __LINE__,  NULL);
   memcpy(self->x+1, x, m*sizeof(double));
   memcpy(self->y+1, y, n*sizeof(double));
   for (comp=0; comp<NCOMP; comp++) {
     for(im=0;im<m;im++) {
       for(jn=0;jn<n;jn++) {
	 self->z[comp][im+1][jn+1]=z[comp][im][jn];
       }
     }
     for (j=1; j<=m; j++) {
       sm2_spline(self->y, self->z[comp][j], self->n, 1.0e30, 1.0e30, self->z2[comp][j], err);
       forwardError(*err, __LINE__, NULL);
     }
   }

   return self;
}
#undef __FUNC__

#define __FUNC__ "init_and_copy_Tsqr_vinschter_table"
interTable2Dneq *init_and_copy_Tsqr_vinschter_table_neq(int m, int n, const double *x,  const double *y, const double ***z,
							error **err)
{
   interTable2Dneq *self;
   int comp, im, jn;
   
   self = init_interTable2Dneq(m, n, err);
   forwardError(*err, __LINE__,  NULL);
   memcpy(self->x+1, x, m*sizeof(double));
   memcpy(self->y+1, y, n*sizeof(double));
   for (comp=0; comp<NCOMP; comp++) {
     for(im=0;im<m;im++) {
       for(jn=0;jn<n;jn++) {
	 self->z[comp][im+1][jn+1]=z[comp][im][jn];
	 testErrorRet(!finite(z[comp][im][jn]), ce_overflow, "overflow", *err, __LINE__, NULL);
       }
     }
   }

   return self;
}
#undef __FUNC__

#define __FUNC__ "set_vinschter"
void set_vinschter(cosmo* self, int m, int n, const double *x,  const double *y, const double ***z, error **err)
{
   printf("set_vinschter\n");
   self->tsqr_vinschter = init_and_copy_Tsqr_vinschter_table_neq(m, n, x, y, z, err);
   forwardError(*err,__LINE__,);
   set_norm(self, err);
   forwardError(*err,__LINE__,);
}
#undef __FUNC__

#define __FUNC__ "Tsqr_vinschter"
double Tsqr_vinschter(cosmo *self, comp_t comp, double k, double a, error **err)
{
   double res;

   testErrorRet(comp!=comp_c&&comp!=comp_b&&comp!=comp_nu, ce_unknown, "unknown species", *err, __LINE__, 0);

   if (self->transfer!=camb_vinschter) {
      *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
      return 0;
   }
   if (self->tsqr_vinschter==NULL) {
      /* Initialization of the table has to be done external of smith2.c */
      *err = addError(ce_no_init, "Transfer function array not initialized, CAMB output not read", *err, __LINE__);
      return 0;
   }

   res = sm2_interpol2Dneq(self->tsqr_vinschter, comp, k, a, err);
   forwardError(*err, __LINE__, 0);

   testErrorRet(!finite(res), ce_overflow, "overflow", *err, __LINE__, 0);
   testErrorRet(res<0.0, ce_negative, "negative", *err, __LINE__, 0);

   return res;
}
#undef __FUNC__

/* ============================================================ *
 * dfridr.c                                                     *
 * NR page 188. Returns derivate of func at x, initial step is  *
 * h. Error estimate in err.                                *
 * Modified! func depends on two double! (like P_L)             *
 * ============================================================ */

#define CON 1.4
#define CON2 (CON*CON)
#define BIG 1.0e30
#define NTAB 10
#define SAFE 2.0
#define __FUNC__ "sm2_dfridr"
double sm2_dfridr(double (*func)(cosmo*,double,double,error **), double x, double h,
                  double *errn, double aa, cosmo* self, error **err)
{
   int i,j;
   double errt,fac,hh,**a,ans;

   ans = 1e30; /* dummy initialization */
   testErrorRet(h==0.0, ce_wrongValue, "h has to be larger than zero", *err, __LINE__, 0.0);
   a=sm2_matrix(1,NTAB,1,NTAB,err);
   forwardError(*err,__LINE__,0);

   hh=h;
   a[1][1]=(*func)(self,aa,x+hh,err);
   forwardError(*err,__LINE__,0);
   a[1][1]=(a[1][1]-(*func)(self,aa,x-hh,err))/(2.0*hh);
   forwardError(*err,__LINE__,0);

   *errn=BIG;
   for (i=2;i<=NTAB;i++) {
      hh /= CON;
      a[1][i]=(*func)(self,aa,x+hh,err);
      forwardError(*err,__LINE__,0);
      a[1][i]=(a[1][i]-(*func)(self,aa,x-hh,err))/(2.0*hh);
      forwardError(*err,__LINE__,0);
      fac=CON2;
      for (j=2;j<=i;j++) {
         a[j][i]=(a[j-1][i]*fac-a[j-1][i-1])/(fac-1.0);
         fac=CON2*fac;
         errt=fmax(fabs(a[j][i]-a[j-1][i]),fabs(a[j][i]-a[j-1][i-1]));
         if (errt <= *errn) {
            *errn=errt;
            ans=a[j][i];
         }
      }
      if (fabs(a[i][i]-a[i-1][i-1]) >= SAFE*(*errn)) break;
   }
   sm2_free_matrix(a,1,NTAB,1,NTAB);
   return ans;
}
#undef CON
#undef CON2
#undef BIG
#undef NTAB
#undef SAFE
#undef __FUNC__

double W_tophat(double x)
{
   if (x<epsilon) {
      double xsqr;
      xsqr = x*x;
      return 1.0 - xsqr/10.0 + xsqr*xsqr/280.0;
   }
   return 3./(x*x*x)*(sin(x) - x*cos(x));
}

#define __FUNC__ "int_for_sigma_R"
double int_for_sigma_R(double logk, void *intpar, error **err)
{
   double res, k, x, R, tt, Omega_m_tot=0.0, f_b=0.0, f_c=0.0, f_nu=0.0, p_c=0.0, p_b=0.0, p_nu=0.0;
   cosmoANDdouble *extra;
   cosmo *self;
   
   k     = exp(logk);
   extra = (cosmoANDdouble*)intpar;
   self  = extra->self;
   R     = extra->r;
   x     = W_tophat(k*R)/3.0;

   switch (self->transfer) {
      case bbks : case eisenhu :
	 /* unnormalized power spectrum */
	 tt = Tsqr(self,k,err);      forwardError(*err, __LINE__, 0);
	 break;
      case camb_vinschter :
	 Omega_m_tot = self->Omega_m + self->Omega_nu_mass;
	 f_b  = self->Omega_b/Omega_m_tot;
	 f_c  = (self->Omega_m-self->Omega_b)/Omega_m_tot;
	 f_nu = self->Omega_nu_mass/Omega_m_tot;
	 p_b  = Tsqr_vinschter(self, comp_b, k, 1.0, err);    forwardError(*err, __LINE__, 0);
	 p_c  = Tsqr_vinschter(self, comp_c, k, 1.0, err);    forwardError(*err, __LINE__, 0);
	 p_nu = Tsqr_vinschter(self, comp_nu, k, 1.0, err);   forwardError(*err, __LINE__, 0);

	 /* NEW: now with p_nu */
	 tt   = dsqr(f_b*sqrt(p_b) + f_c*sqrt(p_c) + f_nu*sqrt(p_nu));
	 //tt   = 1.0/((f_b+f_c)*(f_b+f_c))*dsqr(f_b*sqrt(p_b) + f_c*sqrt(p_c));
	 break;
      default :
	 *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
	 return 0;
   }

   res  = k*dsqr(k)*tt*x*x;
   return res;
}
#undef __FUNC__

int change_sigma_8_sqr(cosmo* avant, cosmo* apres)
{
   return change_Tsqr(avant,apres);
}

/* PD2 42, for a=1, so D_+=1. Normalization of the power spectrum */
#define __FUNC__ "sigma_8_sqr"
double sigma_8_sqr(cosmo* self, error **err)
{
   double integral;
   cosmoANDdouble cANDd;

   if (self->cmp_sigma8<0) {
      cANDd.self = self;
      cANDd.r    = 8.0;    /* [Mpc/h] */
      integral = sm2_qromberg(int_for_sigma_R, (void*)&cANDd, log(k_min), log(k_max), 1.0e-6, err);
      forwardError(*err,__LINE__,0);
      self->cmp_sigma8 = 4.5/pi_sqr*integral;
   }
   if (self->cmp_sigma8<0) {
      *err = addError(ce_negative,"sigma8 can't be negative",*err,__LINE__);
      return 0;
   }
   return self->cmp_sigma8;
}
#undef __FUNC__

#define __FUNC__ "P_L"
double P_L(cosmo* self, double a, double k, error **err)
{
   double d, t, s, ddtt, Omega_m_tot=0.0, f_b=0.0, f_c=0.0, p_c=0.0, p_b=0.0;

   if (k<0) {
      *err = addError(ce_negative,"k Negative !",*err,__LINE__);
      return 0;
   }

   switch (self->transfer) {
      case bbks : case eisenhu :
	 d    = D_plus(self, a, 1, err);          forwardError(*err,__LINE__,0);
	 t    = Tsqr(self, k, err);               forwardError(*err,__LINE__,0);
	 ddtt = d*d*t;
	 break;
      case camb_vinschter :
	 /* CDM+baryons */
	 Omega_m_tot = self->Omega_m + self->Omega_nu_mass;
	 f_b  = self->Omega_b/Omega_m_tot;
	 f_c  = (self->Omega_m-self->Omega_b)/Omega_m_tot;
	 /* Hannestad et al. 2006 (3.4, 3.5) */
	 p_c  = Tsqr_vinschter(self, comp_c, k, a, err);  forwardError(*err, __LINE__, 0);
	 p_b  = Tsqr_vinschter(self, comp_b, k, a, err);  forwardError(*err, __LINE__, 0);
	 ddtt = 1.0/((f_b+f_c)*(f_b+f_c))*dsqr(f_b*sqrt(p_b) + f_c*sqrt(p_c));
	 break;
      default :
	 *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
	 return 0;
   }

   s = sigma_8_sqr(self, err);           forwardError(*err,__LINE__,0);
   return dsqr(self->sigma_8)*ddtt/s;
}
#undef __FUNC__

#define __FUNC__ "P_nu"
double P_nu(cosmo *self, double a, double k, error **err)
{
   double tt_nu;
   tt_nu = Tsqr_vinschter(self, comp_nu, k, a, err);          forwardError(*err, __LINE__, 0);
   tt_nu = tt_nu*dsqr(self->sigma_8)/sigma_8_sqr(self, err);     forwardError(*err, __LINE__, 0);
   return tt_nu;
}
#undef __FUNC__

#define __FUNC__ "n_L"
/* PD 22 */
double n_L(cosmo* self,double a, double k, error **err)
{
   double diff, hh, errn, n;

   hh   = k/20.0;
   diff = sm2_dfridr(P_L, 0.5*k, hh, &errn, a, self, err);
   forwardError(*err,__LINE__,0);
   n    = 0.5*k/P_L(self,a, 0.5*k, err)*diff;
   forwardError(*err,__LINE__,0);
   return n;
}
#undef __FUNC__

#define __FUNC__ "f_NL"
/* PD 21, 23-27 */
double f_NL(cosmo* self, double x, double a, double k, error **err)
{
   double A, B, alpha, beta, V;
   double c, gg, f0, f1, f2, f3, f4;

   c = 1.0 + n_L(self,a,k,err)/3.0;
   forwardError(*err,__LINE__,0);
   testErrorRet(c<0, ce_negative, "Spectral index n_spec too small for PD fitting formula",
		*err, __LINE__, 0);

   A = 0.482/pow(c, 0.947);
   B = 0.226/pow(c, 1.778);
   alpha = 3.31/pow(c, 0.244);
   beta = 0.862/pow(c, 0.287);
   V = 11.55/pow(c, 0.423);

   switch (self->transfer) {
      case bbks : case eisenhu : case camb :
	 gg = D_plus(self, a, 0, err)/a;
	 forwardError(*err,__LINE__,0);
	 break;
      case camb_vinschter :
	 /* TODO: T(K,a), for CDM+b, should be independent of k */
	 *err = addError(ce_transfer, "transfer=camb_vinschter not yet implemented with nonlinear=pd96", *err, __LINE__);
	 return 0.0;
      default :
	 *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
	 return 0;
   }
   
   f0 = pow(A*x, alpha);
   f1 = 1.0 + B*beta*x + pow(f0, beta);
   f2 = f0*gg*gg*gg/(V*sqrt(x));
   f3 = 1.0 + pow(f2, beta);
   f4 = x*pow(f1/f3, 1.0/beta);
	 
   if (!finite(f4)) {
      *err = addError(ce_infnan, "inf or nan encountered", *err, __LINE__);
      return 0;
   }
   return f4;
}
#undef __FUNC__

#define __FUNC__ "sm2_transfer"
double sm2_transfer(cosmo* self, double k, double a, error **err)
{
   double keff, r, Gamma, res, t, teff, Omega_m_tot=0.0, f_b=0.0, f_c=0.0, p_c=0.0, p_b=0.0;

   Gamma = self->h_100*self->Omega_m*exp(-self->Omega_b - sqrt(self->h_100/0.5)
					 *self->Omega_b/self->Omega_m);
   keff = 0.172 + 0.011*dsqr(log(Gamma/0.36));
   r    = k/keff;

   switch (self->transfer) {
      case bbks : case eisenhu : case camb :
	 t    = Tsqr(self,k,err);           forwardError(*err,__LINE__,0);
	 teff = Tsqr(self, keff, err);      forwardError(*err, __LINE__, 0);
	 break;
      case camb_vinschter :
	 Omega_m_tot = self->Omega_m + self->Omega_nu_mass;
	 f_b  = self->Omega_b/Omega_m_tot;
	 f_c  = (self->Omega_m-self->Omega_b)/Omega_m_tot;

	 p_c  = Tsqr_vinschter(self, comp_c, k, a, err);     forwardError(*err, __LINE__, 0);
	 p_b  = Tsqr_vinschter(self, comp_b, k, a, err);     forwardError(*err, __LINE__, 0);
	 t    = 1.0/((f_b+f_c)*(f_b+f_c))*dsqr(f_b*sqrt(p_b) + f_c*sqrt(p_c));

	 p_c  = Tsqr_vinschter(self, comp_c, keff, a, err);     forwardError(*err, __LINE__, 0);
	 p_b  = Tsqr_vinschter(self, comp_b, keff, a, err);     forwardError(*err, __LINE__, 0);
	 teff = 1.0/((f_b+f_c)*(f_b+f_c))*dsqr(f_b*sqrt(p_b) + f_c*sqrt(p_c));

	 break;
      default :
	 *err = addError(ce_transfer, "unknown transfer type", *err, __LINE__);
	 return 0;
   }
   res  = t/teff*r*r*r*dsqr(self->sigma_8);
   testErrorRet(!finite(res), ce_overflow, "Transfer function not finite", *err, __LINE__, 0);

   return res;
}
#undef __FUNC__

/* ============================================================ *
 * Bond&Efstathiou 1984 approximation to Delta_L		*
 * ============================================================ */

#define N_kk 500
/* in units of H_0/c */
#define kmin 1.0e-5
#define kmax 6000.0
int change_Delta_L_BE2(cosmo* avant, cosmo* apres)
{
   if (NCOCLOSE(avant,apres,Omega_m) || NCOCLOSE(avant,apres,h_100) || NCOCLOSE(avant,apres,Omega_b)
       || NCOCLOSE(avant,apres,n_spec) || NCOCLOSE(avant,apres,normalization)
       || NCOCLOSE(avant,apres,normalization))
     return 1;
   return 0;
}

#define __FUNC__ "Delta_L_BE2"  
double Delta_L_BE2(cosmo* self,double k,error **err)
{
   double keff, q8=0.0, logkmin, logkmax, dlogk, tk8=0.0, Gamma=0.0;
   double q, tk;
   double klog, kk, res;
   int j;
   double* table; interTable *transferBE;

   if (self->transfer==camb_vinschter || self->transfer==camb) {
      *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
      return 0;
   }

   if (self->transferBE==NULL) {
      Gamma      = self->h_100*self->Omega_m*exp(-self->Omega_b - sqrt(self->h_100/0.5)*self->Omega_b/self->Omega_m);
      keff       = 0.172 + 0.011*dsqr(log(Gamma/0.36));
      q8         = 1.0e-20 + keff/Gamma;
      tk8        = 1.0/pow(1+pow(6.4*q8+pow(3.0*q8,1.5)+dsqr(1.7*q8),1.13),1.0/1.13);
      logkmin    = log(kmin);
      logkmax    = log(kmax);
      dlogk      = (logkmax - logkmin)/(N_kk-1.);
      transferBE = init_interTable(N_kk, logkmin, logkmax, dlogk, 0.0, 0.0,err);
      forwardError(*err,__LINE__,0);
		
      table = transferBE->table;
      for (j=0,klog=logkmin; j<N_kk; j++,klog+=dlogk) {
	 kk = exp(klog);
	 q  = 1e-20 + kk/Gamma;
	 tk = 1/pow(1+pow(6.4*q+pow(3.0*q,1.5)+dsqr(1.7*q),1.13),1/1.13);
	 table[j] = (3+self->n_spec)*log(q/q8) + 2*log(self->sigma_8*tk/tk8);
      }
      self->transferBE = transferBE;
   }

   kk = log(k);
   if (kk<self->transferBE->a || kk>self->transferBE->b) {
      q  = 1e-20 + k/Gamma;
      tk = 1/pow(1+pow(6.4*q+pow(3.0*q,1.5)+dsqr(1.7*q),1.13),1/1.13);
      res = self->sigma_8*self->sigma_8*pow(q/q8,3+self->n_spec)*tk*tk/tk8/tk8;
   } else {
      res = exp(interpol_wr(self->transferBE, kk,err));
      forwardError(*err,__LINE__,0);
   }
   return res;
}
#undef N_kk
#undef kmin
#undef kmax
#undef __FUNC__

/* ============================================================ *
 * Calculates k_NL, n_eff, n_cur.				*
 * ============================================================ */ 
#define __FUNC__ "int_for_wint2_knl"
double int_for_wint2_knl(double logk, void *intpar, error **err)
{
   double krsqr, k, r, a;
   cosmoAND2double* extra;

   extra = (cosmoAND2double*)intpar;
   r     = extra->r;
   a     = extra->a;
   k     = exp(logk);
   krsqr = dsqr(k*r);
   r     = sm2_transfer(extra->self,k,1.0,err)*exp(-krsqr);
   forwardError(*err,__LINE__,0);

   return r;
}
#undef __FUNC__

#define __FUNC__ "int_for_wint2_neff"
double int_for_wint2_neff(double logk, void *intpar, error **err)
{
   double krsqr, k, r, a;
   cosmoAND2double* extra;

   extra = (cosmoAND2double*) intpar;
   r     = extra->r; 
   a     = extra->a;
   k     = exp(logk);
   krsqr = dsqr(k*r);
   r     = sm2_transfer(extra->self,k,1.0,err)*2.0*krsqr*exp(-krsqr);
   forwardError(*err,__LINE__,0);
   return r;
}
#undef __FUNC__

#define __FUNC__ "int_for_wint2_ncur"
double int_for_wint2_ncur(double logk, void *intpar,error **err)
{
   double krsqr, k, r, a;
   cosmoAND2double* extra;
	
   extra = (cosmoAND2double*) intpar;
   r     = extra->r;
   a     = extra->a;
   k     = exp(logk);
   krsqr = dsqr(k*r);
   r     = sm2_transfer(extra->self,k,1.0,err)*4.0*krsqr*(1.0-krsqr)*exp(-krsqr);
   forwardError(*err,__LINE__,0);
   return r;
}
#undef __FUNC__

#define kmin 1.e-2
#define kmaxdefault 1.e-8
#define __FUNC__ "wint2"
void wint2(cosmo* self, double r, double *sig, double *d1, double *d2, double a, int onlysig,
	   error **err, double precision)
{
   double kmax, logkmin, logkmax, s1, s2, s3, amp;
   cosmoAND2double intpar;

   /* choose upper integration limit to where filter function dropped
    * substantially */
   kmax  = sqrt(10.0*log(10.0))/r;
   if (kmax<kmaxdefault) kmax = kmaxdefault;

   logkmin = log(kmin);
   logkmax = log(kmax);
   intpar.r = r;
   intpar.a = a;                 /* only needed for camb_vinschter */
   intpar.self = self;

   switch (self->transfer) {
      case bbks : case eisenhu : case camb :
	 amp = D_plus(self, a, 1, err);   forwardError(*err,__LINE__,);
	 break;
      case camb_vinschter :
	 amp = 1.0;    /* growth in sm2_transfer */
	 break;
      default :
	 *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
	 return;
   }

   if (onlysig==1) {
      s1   = sm2_qromberg(&int_for_wint2_knl, (void*)&intpar, logkmin, logkmax, precision, err);
      forwardError(*err, __LINE__,);
      *sig = amp*sqrt(s1);
   } else s1 = dsqr(1.0/amp);   /* sigma = 1 */

   if (onlysig==0) {
      s2  = sm2_qromberg(int_for_wint2_neff, (void*)&intpar, logkmin, logkmax, 1.0e-6, err);
      forwardError(*err, __LINE__,);
      s3  = sm2_qromberg(int_for_wint2_ncur, (void*)&intpar, logkmin, logkmax, 1.0e-6, err);
      forwardError(*err, __LINE__,);
      *d1 = -s2/s1;
      *d2 = -dsqr(*d1) - s3/s1;
   }
}
#undef kmin
#undef kmaxdefault
#undef __FUNC__

/* slope in the highly nonlinear regime, c.f. Smith et al (2003) eq. (61) */
#define __FUNC__ "slope_NL"
double slope_NL(double n, double ncur, double om_m, double om_v, error **err)
{
   double gam, f1a, f1b, frac, f1;

   gam = 0.86485 + 0.2989*n + 0.1631*ncur;
   if(fabs(1-om_m)>0.01) {
      f1a = pow(om_m,(-0.0732));
      f1b = pow(om_m,(-0.0307));
      frac = om_v/(1.0-om_m);  
      f1 = frac*f1b + (1-frac)*f1a;
   } else {
      f1 = 1.0;
   }

   return 3.0*(f1-1.0) + gam - 3.0;
}
#undef __FUNC__

#define __FUNC__ "halofit"
void halofit(double k, double n, double ncur, double knl, double plin, 
	     double om_m, double om_v, double *pnl, error **err)
{
   double gam,a,b,c,xmu,xnu,alpha,beta,f1,f2,f3;
   double y, ysqr;
   double f1a,f2a,f3a,f1b,f2b,f3b,frac,pq,ph;
   double nsqr;

   nsqr  = n*n;
   gam   = 0.86485 + 0.2989*n + 0.1631*ncur;
   a     = 1.4861 + 1.83693*n + 1.67618*nsqr + 0.7940*n*nsqr
           + 0.1670756*nsqr*nsqr - 0.620695*ncur;
   a     = pow(10,a);
   b     = pow(10,(0.9463+0.9466*n+0.3084*nsqr-0.940*ncur));
   c     = pow(10,(-0.2807+0.6669*n+0.3214*nsqr-0.0793*ncur));
   xmu   = pow(10,(-3.54419+0.19086*n));
   xnu   = pow(10,(0.95897+1.2857*n));
   alpha = 1.38848+0.3701*n-0.1452*nsqr;
   beta  = 0.8291+0.9854*n+0.3400*nsqr;

   if(fabs(1-om_m)>0.01) {
      f1a = pow(om_m,(-0.0732));
      f2a = pow(om_m,(-0.1423));
      f3a = pow(om_m,(0.0725));
      f1b = pow(om_m,(-0.0307));
      f2b = pow(om_m,(-0.0585));
      f3b = pow(om_m,(0.0743));
  
      frac = om_v/(1.0-om_m);  

      //new - interpolation in eq. of state parameters
      /*
      double we, wde;
      // aa has to be given as argument to halofit
      wde = w_de(self, aa, err);
      forwardError(*err, __LINE__,);
      // Interpolate between this w and w=-1/3 (behaves like OCDM)
      we   = frac*w_de + (1.0-frac)*(-1.0/3.0);
      frac = -(3.0*we+1.0)/2.0;
      //end of new
      */

      f1 = frac*f1b + (1-frac)*f1a;
      f2 = frac*f2b + (1-frac)*f2a;
      f3 = frac*f3b + (1-frac)*f3a;
   } else {      /* EdS Universe */
      f1 = f2 = f3 = 1.0;
   }

   y = k/knl;
   ysqr = y*y;
   ph = a*pow(y,f1*3)/(1+b*pow(y,f2)+pow(f3*c*y,3-gam));
   ph = ph/(1+xmu/y+xnu/ysqr);
   pq = plin*pow(1+plin,beta)/(1+plin*alpha)*exp(-y/4.0-ysqr/8.0);
   *pnl = pq + ph;

   testErrorRet(!finite(*pnl), ce_overflow, "Power spectrum pnl not finite", *err, __LINE__,);
   testErrorRet(*pnl<0, ce_negative,
		"Negative power spectrum pnl. This can happen if n_eff is very negative and plin large.",
		*err, __LINE__,);
}
#undef __FUNC__
 
#define __FUNC__ "dlog"
double dlog(double x)
{
   return log(x)/log(10.0);
}
#undef __FUNC__

int change_P_NL(cosmo* avant, cosmo* apres) {
   if (NCOEQ(avant,apres,growth) || NCOEQ(avant,apres,transfer) || NCOEQ(avant,apres,de_param) || 
       NCOEQ(avant,apres,nonlinear) || NCOEQ(avant,apres,N_a))
     return 1;
   if (NCOCLOSE(avant,apres,Omega_m) || NCOCLOSE(avant,apres,Omega_de) || NCOCLOSE(avant,apres,w0_de)||
       NCOCLOSE(avant,apres,w1_de) || NCOCLOSE(avant,apres,a_min) || NCOEQ(avant,apres,de_param) || 
       NCOCLOSE(avant,apres,h_100) || NCOCLOSE(avant,apres,Omega_b) || NCOCLOSE(avant,apres,n_spec)||
       NCOCLOSE(avant,apres,normalization) || 
       NCOCLOSE(avant,apres,Omega_nu_mass) || NCOCLOSE(avant,apres,Neff_nu_mass))
     return 1;
   return 0;
}

/* ============================================================ *
 * Dark (+ baryonic) matter non-linear power spectrum		*
 * ============================================================ */
#define itermax       25
#define logrmin       -6.0
#define logrmax       5.0
#define eps_a         1.0e-5
#define epsi          1.0e-3
#define precision_min 1.0e-6
#define precision_max 1.01e-11
#define __FUNC__ "P_NL"
double P_NL(cosmo* self, double a, double k_NL, error **err)
{
   double table_k[N_k], table_P[N_k], y2[N_k];
   double *table_slope = 0;
   double logkmin = 0.0, logkmax = 0.0, dk = 0.0, da = 0.0;
   double Delta_NL, Delta_L, k_L, lnk_NL=0.0;
   double precision = precision_min;

   double omm=0.0, omv=0.0, amp=0.0;
   double logr1, logr2 ,diff, rmid, sig, d1, d2;
   double rknl=0.0, rneff=0.0, rncur=0.0;

   double aa, klog, val, logrmid, EE, r, upper=0.0;
   int i,j, iter, golinear=-1;
   interTable2D *lP_NL;

   testErrorRetVA(k_NL<=0, ce_negative, "Negative k=%g", *err, __LINE__, -1.0, k_NL);
   testErrorRetVA(a<self->a_min || a>1, ce_wrongValue, "a=%g out of range", *err, __LINE__, -1.0, a);
		
   if (self->P_NL == NULL) {
      testErrorRet(!(self->nonlinear==linear || self->nonlinear==pd96 || self->nonlinear==smith03),
		   ce_unknown,
		   "Unknown nonlinear mode, has to be one of \'linear\', \'pd96\' or \'smith03\'",
		   *err, __LINE__, 0);

      /* upper = (dlnP/dlnk)_{k=kmax}, for splines & extrapolation */
      /* Note that the in the range considered here the linear power
       * spectrum is still a bit shallower than k!(n-4) since T(k) has
       * not yet reached its asymptotic limit. 
       */

      da = (1.0 - self->a_min)/(self->N_a-1.0);
      aa = self->a_min;
      logkmin = log(k_min);
      logkmax = log(k_max);
      dk = (logkmax - logkmin)/(N_k-1.0);
		 
      if (self->nonlinear==linear) {
	 upper = self->n_spec-4.0;
      } else if (self->nonlinear==pd96) {
	 upper = -2.5;
      } else if (self->nonlinear==smith03) {
	 self->slope = init_interTable(self->N_a, self->a_min, 1.0, da, 1.0e31, 1.0e31, err);
	 forwardError(*err,__LINE__,0);
	 table_slope = self->slope->table;
      }
      lP_NL = init_interTable2D(self->N_a, self->a_min, 1.0, da, N_k, logkmin, logkmax, 
				dk, self->n_spec, upper, err);
      forwardError(*err,__LINE__,0);
		 
      for (i=0; i<self->N_a; i++,aa+=da) {
			 
	 if (self->nonlinear==smith03) {           /* Smith et al. (2003) */
	    EE = Esqr(self, aa, 0, err);           forwardError(*err,__LINE__,0);
	    omm = Omega_m_a(self, aa, EE, err);    forwardError(*err,__LINE__,0);
	    omv = Omega_de_a(self, aa, EE, err);   forwardError(*err,__LINE__,0);
	    golinear = 0;
				 
	    /* find non-linear scale with iterative bisection */
	 iter_start:
	    logr1 = logrmin;
	    logr2 = logrmax;
	    iter = 0;
	    do {
					 
	       logrmid = (logr2+logr1)/2.0;
	       rmid    = pow(10,logrmid);
	       wint2(self,rmid, &sig, NULL, NULL, aa, 1, err, precision);
	       if (isError(*err))
		 printError(stderr, *err);
	       forwardError(*err, __LINE__, 0);
	       diff = sig - 1.0;
	       if(diff>epsi)
		 logr1 = dlog(rmid);
	       if(diff<-epsi)
		 logr2 = dlog(rmid);
	    } while (fabs(diff)>=epsi && ++iter<itermax);
				 
	    /* non-linear scale not found */
	    if (iter>=itermax) {
	       /* non-linear scale very very small, all scales basically
		* linear: set flag golinear and continue
		*/
	       if (logrmid-logrmin<epsilon) {
		  /* scale at low end */
		  golinear = 1;
		  upper = table_slope[i] = self->n_spec-4.0;
	       } else if (logrmax-logrmid<epsilon) {
		  /* scale at high end */
		  *err = addError(ce_nonlin, "All scales are nonlinear. Something went wrong in P_NL",
				  *err,__LINE__);
		  return 0.0;
	       } else {
		  /* somewhere in between */
		  if (precision>precision_max) {
		     precision /= 10.0;
		     goto iter_start;
		  } else {
		     printf("a=%f %f %f %e\n", aa, logrmid, logrmin, precision);
		     *err = addError(ce_noknl, "Nonlinear scale not found", *err, __LINE__);
		     return 0.0;
		  }
	       }
	    } else {
	       /* spectral index & curvature at non-linear scale */
	       wint2(self,rmid, &sig, &d1, &d2, aa, 0, err, precision_min);
	       forwardError(*err,__LINE__,0);
	       rknl  = 1.0/rmid;
	       rneff = -3.0-d1;
	       rncur = -d2;
	       upper = table_slope[i] = slope_NL(rneff, rncur, omm, omv, err);
	       forwardError(*err,__LINE__,0);
	    }
	 }
	 klog = logkmin;
	 for (j=0; j<N_k; j++, klog+=dk) {
	    k_L = exp(klog);
	    if (self->nonlinear==linear) {

	       Delta_NL = P_L(self,aa,k_L,err)*k_L*k_L*k_L/(2.0*pi_sqr);
	       forwardError(*err,__LINE__,0);
	       lnk_NL   = klog;

	    } else if (self->nonlinear==pd96) {

	       Delta_L  = P_L(self,aa,k_L,err)*k_L*k_L*k_L/(2.0*pi_sqr);
	       forwardError(*err,__LINE__,0);
	       Delta_NL = f_NL(self,Delta_L, aa, k_L, err);
	       forwardError(*err,__LINE__,0);
	       lnk_NL   = klog + 1.0/3.0*log(1.0 + Delta_NL);  /* PD (5) */

	    } else if (self->nonlinear==smith03) {

	       switch (self->transfer) {
		  case bbks : case eisenhu : case camb :
		     amp = D_plus(self, aa, 1, err);        forwardError(*err,__LINE__,0);
		     break;
		  case camb_vinschter :
		     amp = 1.0;  /* growth in Tsqr_cambvinschter */
		     break;
		  default :
		     *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
		     return 0;
	       }

	       Delta_L = amp*amp*sm2_transfer(self, k_L, aa, err);
	       forwardError(*err, __LINE__, 0);
	       if (golinear==0) {
		  halofit(k_L, rneff, rncur, rknl, Delta_L, omm, omv, &Delta_NL, err);
		  forwardError(*err, __LINE__, 0);

	       } else if (golinear==1) {
		  Delta_NL = Delta_L;
	       }
	       lnk_NL = klog;

	    }
				 
	    table_k[j] = lnk_NL;
	    table_P[j] = log(2*pi_sqr*Delta_NL) - 3.0*lnk_NL; /* PD (3) */
	 }
	 sm2_spline(table_k-1, table_P-1, N_k, self->n_spec, upper, y2-1,err);
	 forwardError(*err,__LINE__,0);
	 klog = logkmin;
	 for (j=0; j<N_k; j++, klog += dk) {
	    sm2_splint(table_k-1, table_P-1, y2-1, N_k, klog, &val,err);
	    forwardError(*err,__LINE__,0);
	    lP_NL->table[i][j] = val;
	 }
      }
      self->P_NL = lP_NL;
   }

   klog = log(k_NL);

#ifdef fastxi
   if (self->nonlinear==smith03) {
      upper = interpol_wr(self->slope,a,err);
      forwardError(*err, __LINE__, 0);
      self->P_NL->upper = upper;
   }

   r = (a-self->a_min)/self->P_NL->dx1;   /* da */
   i = (int)(r+0.5);                      /* round to next integer */
   if (fabs(r-i)<eps_a) {
      val = sm2_interpol(self->P_NL->table[i], self->P_NL->n2, self->P_NL->a2, self->P_NL->b2,
			 self->P_NL->dx2, klog, self->n_spec, self->P_NL->upper, err);
      forwardError(*err, __LINE__, 0);
      testErrorRet(!finite(val), ce_infnan, "Power spectrum P_NL not finite", *err, __LINE__, -1);
      val = exp(val);
      return val;
   }
   else
#endif
   val = interpol2D(self->P_NL, a, klog, err);
   forwardError(*err, __LINE__, 0);
   val = exp(val);
   return val;
}
#undef itermax 
#undef logrmin 
#undef logrmax 
#undef eps_a  
#undef epsi
#undef precision_min
#undef precision_max
#undef __FUNC__

/* ============================================================ *
 * Total (dark+baryon+neutrino) power spectrum			*
 * ============================================================ */
#define __FUNC__ "P_NL_tot"
double P_NL_tot(cosmo *self, double a, double k, error **err)
{
   double p_cb, p_nu, p_cb_nl, p_tot=0.0, f_nu, f_bc, Omega_m_tot;

   /* matter power spectrum */
   if (self->nonlinear==linear){
      p_cb = P_L(self, a, k, err);
      forwardError(*err, __LINE__, 0);
   } else if (self->nonlinear==pd96||self->nonlinear==smith03) {
      p_cb = P_NL(self, a, k, err);
      forwardError(*err, __LINE__, 0);
   } else {
      *err = addErrorVA(ce_unknown, "Unknown nonlinear flag %d", *err, __LINE__, self->nonlinear);
      return -1.0;
   }
   forwardError(*err, __LINE__, 0);

   switch (self->transfer) {
      case bbks : case eisenhu :
	 p_tot = p_cb;
	 break;

      case camb_vinschter :
	 /* Linear neutrino power spectrum */

	 Omega_m_tot = self->Omega_m + self->Omega_nu_mass;
	 testErrorRet(Omega_m_tot<0.0, ce_negative, "Omega_m + Omega_nu_mass < 0", *err, __LINE__, 0.0);

	 f_nu = self->Omega_nu_mass/Omega_m_tot;
	 f_bc = self->Omega_m/Omega_m_tot;
	 /* Hannestad et al. 2006 (3.4, 3.5) */
	 p_cb_nl = p_cb;
	 p_nu = P_nu(self, a, k, err);                         forwardError(*err, __LINE__, 0);
	 p_tot = dsqr(f_nu*sqrt(p_nu) + f_bc*sqrt(p_cb_nl));
	 break;

      default :
	 *err = addError(ce_transfer, "wrong transfer type", *err, __LINE__);
	 return 0;
   }

   return p_tot;
}
#undef __FUNC__

#define __FUNC__ "int_for_w"
double int_for_w(double a, void *intpar, error **err)
{
   double asqr, d;
   cosmoANDint *ci;

   ci = (cosmoANDint*)intpar;
   asqr = dsqr(a);
   d    = asqr*asqr*Esqr(ci->self, a, ci->i, err);
   forwardError(*err,__LINE__,-1);

   return 1.0/sqrt(d);
}
#undef __FUNC__

int change_w(cosmo *avant, cosmo *apres) {
   if (NCOEQ(avant,apres,de_param) || NCOEQ(avant,apres,N_a)) 
     return 1;
   if (NCOCLOSE(avant,apres,Omega_m) || NCOCLOSE(avant,apres,Omega_de) || NCOCLOSE(avant,apres,w0_de) ||
       NCOCLOSE(avant,apres,w1_de) || NCOCLOSE(avant,apres,a_min) || NCOCLOSE(avant,apres,Omega_nu_mass))
     return 1;
   return 0;
}

/* Comoving distance to an object at scale factor a */
/* BS01 2.41, with a(z_2) = 0, a(z_1) = z, [w] = Mpc/h */
#define __FUNC__ "w"
double w(cosmo* self, double a, int wOmegar, error **err)
{
   double *table;
   double da;
   double aa;
   int    i;
   double res;
   interTable *lw;
   cosmoANDint intpar;
	 
   if (self->w==NULL) {
      da = (1.0-self->a_min)/(self->N_a-1.0);
      aa = self->a_min;
      lw = init_interTable(self->N_a, self->a_min, 1.0, da, 0.0, 0.0,err);
      forwardError(*err,__LINE__,0);
      table = lw->table;
      intpar.self = self;
      intpar.i    = wOmegar;
      for (i=0; i<self->N_a-1; i++, aa+=da) {
	 table[i] = R_HUBBLE*sm2_qromberg(int_for_w, (void*)(&intpar), aa, 1.0, 1.0e-6, err);
	 forwardError(*err,__LINE__,0);
      }
      table[self->N_a-1] = 0.0;
      self->w = lw;
   }
   res = interpol_wr(self->w, a, err);
   forwardError(*err,__LINE__,0);
   return res;
}
#undef __FUNC__

/* dw/da in Mpc/h */
#define __FUNC__ "dwoverda"
double dwoverda(cosmo *self, double a, error **err)
{
   double res;
   res = R_HUBBLE*int_for_w(a, (void*)self, err);
   forwardError(*err,__LINE__,0);
   return res;
}
#undef __FUNC__

#define __FUNC__ "f_K"
/* BS01 2.4, 2.30 */
double f_K(cosmo* self, double w, error **err)
{
   double K, K_h, f;

   testErrorRet(w<0, ce_negative, "Comoving distance is negative", *err, __LINE__, -1.0);

   /* Curvature without radiation (photons + massless neutrinos) density */
   K = self->Omega_m + self->Omega_nu_mass + self->Omega_de - 1.0;
   if (K>epsilon) {             /* closed */
      K_h = sqrt(K)/R_HUBBLE;
      testErrorRet(K_h*w>pi, ce_zmax, "z larger than zmax in closed Universe",
		   *err, __LINE__, 0.0);
      f = 1.0/K_h*sin(K_h*w);
   } else if (K<-epsilon) {     /* open */
      K_h = sqrt(-K)/R_HUBBLE;
      f = 1.0/K_h*sinh(K_h*w);
   } else {                     /* flat */
      f = w;
   }

   return f;
}
#undef __FUNC__

/* Luminosity distance in Mpc/h */
#define __FUNC__ "D_lum"
double D_lum(cosmo *self, double a, error **err)
{
   double dlum, ww;
   ww   = w(self, a, 0, err);          forwardError(*err, __LINE__, 0);
   dlum = f_K(self, ww, err)/a;        forwardError(*err, __LINE__, 0);
   return dlum;
}
#undef __FUNC__


