/* ============================================================ *
 * lensing.c							*
 *								*
 * Martin Kilbinger, Karim Benabed 2006, 2007, 2008		*
 * With many thanks to P. Schneider, J. Hartlap and P. Simon.	*
 * ============================================================ */


#include "lensing.h"


#define __FUNC__ "init_parameters_lens"
cosmo_lens *init_parameters_lens(double OMEGAM, double OMEGADE, double W0_DE, double W1_DE, 
				 double H100, double OMEGAB, double OMEGANUMASS,
				 double NEFFNUMASS, double NORM, double NSPEC,
				 int Nzbin, const int *Nnz, const nofz_t *nofz, double *par_nz,
				 nonlinear_t NONLINEAR, transfer_t TRANSFER,
				 growth_t GROWTH, de_param_t DEPARAM,
				 norm_t NORMMODE, tomo_t TOMO, sm2_error **err)
{
   cosmo_lens *res;
   double amin;

   res = malloc_err(sizeof(cosmo_lens), err);  forwardError(*err, __LINE__, NULL);

   res->redshift = init_redshift(Nzbin, Nnz, nofz, par_nz, err); forwardError(*err, __LINE__, NULL);

   amin = get_amin(res->redshift, err);
   forwardError(*err, __LINE__, NULL);

   res->cosmo = init_parameters(OMEGAM, OMEGADE, W0_DE, W1_DE, H100, OMEGAB, OMEGANUMASS,
				NEFFNUMASS, NORM, NSPEC, NONLINEAR, TRANSFER, GROWTH, DEPARAM,
				NORMMODE, amin, err);
   forwardError(*err, __LINE__, NULL);

   /* Lensing parameters */
   res->tomo      = TOMO;
   strcpy(res->stomo, stomo_t(TOMO));

   /* Reset pre-computed tables */
   res->g_source  = NULL;
   res->Pkappa    = NULL;
   res->xiP       = NULL;
   res->xiM       = NULL;
   res->gamma     = NULL;
   res->map_gauss = NULL;
   res->map_poly  = NULL;
 
   return res;
}
#undef __FUNC__

#define __FUNC__ "copy_parameters_lens_only"
cosmo_lens* copy_parameters_lens_only(cosmo_lens* source, sm2_error **err)
{
   cosmo_lens *res;

   res = init_parameters_lens(source->cosmo->Omega_m,source->cosmo->Omega_de, source->cosmo->w0_de,
			      source->cosmo->w1_de, source->cosmo->h_100, source->cosmo->Omega_b,
			      source->cosmo->Omega_nu_mass, source->cosmo->Neff_nu_mass,
			      source->cosmo->normalization, source->cosmo->n_spec,
			      source->redshift->Nzbin, source->redshift->Nnz, source->redshift->nofz,
			      source->redshift->par_nz,
			      source->cosmo->nonlinear, source->cosmo->transfer,
			      source->cosmo->growth, source->cosmo->de_param, 
			      source->cosmo->normmode, source->tomo, err);
   forwardError(*err, __LINE__, NULL);

   return res;
}
#undef __FUNC__

#define __FUNC__ "copy_parameters_lens"
cosmo_lens* copy_parameters_lens(cosmo_lens* source, sm2_error **err)
{
   cosmo_lens *res;
   int Nzbin;

   res = copy_parameters_lens_only(source, err);             forwardError(*err, __LINE__, NULL);

   /* Reset cosmo and redshift */
   free_parameters(&res->cosmo);
   res->cosmo = copy_parameters(source->cosmo, err);         forwardError(*err, __LINE__, NULL);

   free_redshift(&res->redshift);
   res->redshift  = copy_redshift(source->redshift, err);    forwardError(*err, __LINE__, NULL);

   Nzbin = res->redshift->Nzbin;

   res->tomo = source->tomo;
   strcpy(res->stomo, source->stomo);

   res->g_source  = copy_interTable_arr(source->g_source, Nzbin, err);
   forwardError(*err, __LINE__, NULL);
   res->Pkappa = copy_interTable_arr(source->Pkappa, Nzbin*Nzbin, err);
   forwardError(*err, __LINE__, NULL);
   res->xiP    = copy_interTable_arr(source->xiP, Nzbin*Nzbin, err);
   forwardError(*err, __LINE__, NULL);
   res->xiM    = copy_interTable_arr(source->xiM, Nzbin*Nzbin, err);
   forwardError(*err, __LINE__, NULL);
   res->gamma  = copy_interTable_arr(source->gamma, Nzbin*Nzbin, err);
   forwardError(*err, __LINE__, NULL);
   res->map_poly = copy_interTable_arr(source->map_poly, Nzbin*Nzbin, err);
   forwardError(*err, __LINE__, NULL);
   res->map_gauss = copy_interTable_arr(source->map_gauss, Nzbin*Nzbin, err);
   forwardError(*err, __LINE__, NULL);

   return res;
}
#undef __FUNC__

#define __FUNC__ "read_cosmological_parameters_lens"
void read_cosmological_parameters_lens(cosmo_lens **self, FILE *F, error **err)
{
   cosmo_lens *tmp;
   struct { char cosmo_file[128], nofz_file[128]; } tmp2;
   config_element c;
   int j;
   char line[1024];
   FILE *FD;

   tmp = set_cosmological_parameters_to_default_lens(err);
   forwardError(*err, __LINE__,);

   CONFIG_READ_S(&tmp2, cosmo_file, s, F, c, err);
   if (strcmp(tmp2.cosmo_file, "-")!=0) {
      FD = fopen_err(tmp2.cosmo_file, "r", err);
      forwardError(*err, __LINE__,);
   } else {
      FD = F;
   }
   read_cosmological_parameters(&tmp->cosmo, FD, err);
   forwardError(*err, __LINE__,);
   if (strcmp(tmp2.cosmo_file, "-")!=0) fclose(FD);

   CONFIG_READ_S(&tmp2, nofz_file, s, F, c, err);
   if (strcmp(tmp2.nofz_file, "-")!=0) {
      FD = fopen_err(tmp2.nofz_file, "r", err);
      forwardError(*err, __LINE__,);
   } else {
      FD = F;
   }
   read_redshift_info(&(tmp->redshift), FD, err);
   forwardError(*err, __LINE__,);
   if (strcmp(tmp2.cosmo_file, "-")!=0) fclose(FD);

   /* Here: Read lens parameters (currently there are none) */
   CONFIG_READ_S(tmp, stomo, s, F, c, err);
   //STRING2ENUM_SHORT(tmp, tomo, err);
   STRING2ENUM(tmp, tomo, stomo, tomo_t, stomo_t, Ntomo_t, err);
   
   *self = copy_parameters_lens_only(tmp, err);
   forwardError(*err, __LINE__,);
}
#undef __FUNC__

#define __FUNC__ "updateFrom_lens"
void updateFrom_lens(cosmo_lens *avant, cosmo_lens *apres, error **err)
{
   int Nzbin;

   Nzbin = apres->redshift->Nzbin;

   if (change_g_source(avant,apres)) {
      del_interTable_arr(&(apres->g_source), Nzbin);
   }

   if (change_Pkappa(avant,apres)) {
      del_interTable_arr(&(apres->Pkappa), Nzbin*Nzbin);
      del_interTable_arr(&(apres->xiP), Nzbin*Nzbin);
      del_interTable_arr(&(apres->xiM), Nzbin*Nzbin);
      del_interTable_arr(&(apres->gamma), Nzbin*Nzbin);
      del_interTable_arr(&(apres->map_poly), Nzbin*Nzbin);
      del_interTable_arr(&(apres->map_gauss), Nzbin*Nzbin);
   }

   updateFrom(avant->cosmo, apres->cosmo, err);
   forwardError(*err, __LINE__,);

   updateFrom_redshift(avant->redshift, apres->redshift, err);
   forwardError(*err, __LINE__,);
}
#undef __FUNC__

#define NZBIN 1
#define NNZ 5
#define __FUNC__ "set_cosmological_parameters_to_default_lens"
cosmo_lens *set_cosmological_parameters_to_default_lens(error **err)
{
   int    Nnz[NZBIN]        = {NNZ};
   double par_nz[NZBIN*NNZ] = {0.0, 6.0, 0.612, 8.125, 0.62};
   nofz_t nofz[NZBIN]       = {ymmk};

   return init_parameters_lens(0.25, 0.75, -1.0, 0.0, 0.70, 0.044, 0.0, 0.0, 0.80, 1.0,
			       NZBIN, Nnz, nofz, par_nz, smith03, eisenhu, growth_de, linder,
			       norm_s8, tomo_all, err);
}
#undef __FUNC__
#undef NNZ
#undef NZBIN

void free_parameters_lens(cosmo_lens** self)
{
   cosmo_lens *s;
   int Nzbin;

   s = *self;

   Nzbin = s->redshift->Nzbin;
   del_interTable_arr(&s->g_source, Nzbin);

   del_interTable_arr(&s->Pkappa, Nzbin);
   del_interTable_arr(&s->xiP, Nzbin);
   del_interTable_arr(&s->xiM, Nzbin);
   del_interTable_arr(&s->gamma, Nzbin);
   del_interTable_arr(&s->map_poly, Nzbin);
   del_interTable_arr(&s->map_gauss, Nzbin);

   free_redshift(&s->redshift);
   free_parameters(&s->cosmo);

   free(s);
   s = NULL;
}

#define __FUNC__ "dump_param_lens"
void dump_param_lens(cosmo_lens* self, FILE *F, error **err)
{
   dump_param(self->cosmo, F);
   dump_redshift(self->redshift, F, err);
   forwardError(*err, __LINE__,);
   fprintf(F, "# (s)tomo = (%s)(%d)\n", stomo_t(self->tomo), self->tomo);
}
#undef __FUNC__

#define __FUNC__ "int_for_g"
double int_for_g(double aprime, void *intpar, error **err)
{
   cosmo_lensANDintANDdouble* cANDdANDe;
   cosmo_lens* self;
   double ww, wprime, res, fKwp, a;
   int n_bin;
   
   cANDdANDe = (cosmo_lensANDintANDdouble*)intpar;
   self      = cANDdANDe->self;
   a         = cANDdANDe->r;
   n_bin     = cANDdANDe->i;
   
   ww     = w(self->cosmo, a, 0, err);                        forwardError(*err,__LINE__,0);
   wprime = w(self->cosmo, aprime, 0, err);                   forwardError(*err,__LINE__,0);
   res    = prob(self->redshift, 1.0/aprime-1.0, n_bin, err); forwardError(*err,__LINE__,0);
   res   *= f_K(self->cosmo, wprime-ww, err)/dsqr(aprime);    forwardError(*err,__LINE__,0);
   fKwp   = f_K(self->cosmo, wprime, err);                    forwardError(*err,__LINE__,0);

   return res/fKwp;
}
#undef __FUNC__

int change_g_source(cosmo_lens* avant, cosmo_lens* apres)
{
   if (change_w(avant->cosmo, apres->cosmo) || NCOEQ(avant->cosmo, apres->cosmo, N_a))
     return 1;
   if (change_prob(avant->redshift, apres->redshift))
     return 1;
   return 0;
}

/* S98 2.9 */
#define __FUNC__ "g_source"
double g_source(cosmo_lens* self, double a, int n_bin, error **err)
{
   double *table, res;
   double da = 0.0;
   double aa, wdelta, fKz0, wa;
   int    i, nn;
   cosmo_lensANDintANDdouble intpar;
	 
   if (self->redshift->nofz[n_bin]==single) {
      aa = 1.0/(1.0+get_zmin(self->redshift, n_bin));       /* zmin = zmax = z0 */
      if (a<=aa) return 0.0;
      wdelta = w(self->cosmo, aa, 0, err);           forwardError(*err,__LINE__,0);
      wa     = w(self->cosmo, a, 0, err);            forwardError(*err,__LINE__,0);
      res    = f_K(self->cosmo, wdelta-wa, err);     forwardError(*err,__LINE__,0);
      fKz0   = f_K(self->cosmo, wdelta, err);        forwardError(*err,__LINE__,0);

      return res/fKz0;
   }

   if (self->g_source==NULL) {
      da = (1.0-self->cosmo->a_min)/(self->cosmo->N_a-1.0);
      self->g_source = init_interTable_arr(self->redshift->Nzbin, self->cosmo->N_a,
      					  self->cosmo->a_min, 1.0, da, 0.0, 0.0, err);
      forwardError(*err,__LINE__,0);
      for (nn=0; nn<self->redshift->Nzbin; nn++) {
	 table       = self->g_source[nn]->table;
	 table[0]    = 0.0;
	 aa          = self->cosmo->a_min+da;
	 intpar.self = self;
	 intpar.i    = nn;

	 for (i=1;i<self->cosmo->N_a-1;i++,aa+=da) {
	    intpar.r  = aa;
	    table[i]  = sm2_qromberg(int_for_g, (void*)&intpar, self->cosmo->a_min, aa, 1.0e-6, err);
	    forwardError(*err,__LINE__,0);
	 }

	 table[self->cosmo->N_a-1] = 1.0;
      }
   }

   res = interpol_wr(self->g_source[n_bin], a, err);
   forwardError(*err,__LINE__,0);

   return res;
}
#undef __FUNC__

#define __FUNC__ "G"
/* lens efficiency */
double G(cosmo_lens* self, double a, int n_bin, error **err)
{
   double res;
   res = 1.5/dsqr(R_HUBBLE)*(self->cosmo->Omega_m+self->cosmo->Omega_nu_mass)/a
     *g_source(self, a, n_bin, err);
   forwardError(*err, __LINE__, 0);
   return res;	
}
#undef __FUNC__

#define __FUNC__ "int_for_p_2"
double int_for_p_2(double a, void *intpar, error **err)
{
   double hoverh0, asqr, s, fKw, f, res, wa, gg;
   int i_bin, j_bin;
   cosmo_lensANDiid* cANDiid;
   cosmo_lens* self;
  
   cANDiid = (cosmo_lensANDiid*)intpar;
   self    = cANDiid->self;
   s       = cANDiid->r;
   i_bin   = cANDiid->i;
   j_bin   = cANDiid->j;
  
   testErrorRet(a>=1.0, ce_overflow, "Scale factor a>=1", *err, __LINE__, -1);

   asqr    = dsqr(a);
   wa      = w(self->cosmo, a, 0, err);          forwardError(*err, __LINE__, -1);
   fKw     = f_K(self->cosmo, wa, err);          forwardError(*err, __LINE__, -1);
   f       = s/fKw;

   hoverh0 = Esqr(self->cosmo, a, 0, err);       forwardError(*err, __LINE__,-1);
   hoverh0 = sqrt(hoverh0);

   gg   = g_source(self, a, i_bin, err);         forwardError(*err, __LINE__, -1);
   gg  *= g_source(self, a, j_bin, err);         forwardError(*err, __LINE__, -1);
   res  = gg/(asqr*asqr)/hoverh0*R_HUBBLE;
   res *= P_NL_tot(self->cosmo, a, f, err);      forwardError(*err,__LINE__, -1);

   testErrorRet(!finite(res), ce_overflow, "Value not finite", *err, __LINE__, -1);
		
   return res;
}
#undef __FUNC__


int change_Pkappa(cosmo_lens* avant, cosmo_lens* apres)
{
   if (change_w(avant->cosmo, apres->cosmo) || NCOEQ(avant->cosmo, apres->cosmo, N_a)) return 1;
   if (change_Tsqr(avant->cosmo, apres->cosmo)) return 1;
   if (change_prob(avant->redshift, apres->redshift)) return 1;
   if (change_P_NL(avant->cosmo, apres->cosmo)) return 1;
   if (NCOEQ(avant, apres, tomo)) return 1;
   return 0;
}

/* S98  3.4 *
 * i_bin <= j_bin */
#define __FUNC__ "Pkappa"
double Pkappa(cosmo_lens* self, double s, int i_bin, int j_bin, error **err)
{
   double *table;
   double ds, logsmin, logsmax, prefactor;
   double ss, slog, f1, f2;
   int    i, Nzbin, Nzcorr, ii, jj;
   double a, da;
   cosmo_lensANDiid intpar;
	 
   testErrorRetVA(s<=0, ce_negative, "Negative or zero 2d Fourier vector l=%g",* err, __LINE__, -1, s);
   testErrorRet(i_bin>j_bin, lensing_tomoij, "Pkappa_ij defined for i<=j", *err, __LINE__, -1);

   Nzbin  = self->redshift->Nzbin;
   Nzcorr = Nzbin*(Nzbin+1)/2;
   if (self->Pkappa==NULL) {

      logsmin   = log(s_min);
      logsmax   = log(s_max);
      ds        = (logsmax - logsmin)/(N_s - 1.0);
      da        = (1.0 - self->cosmo->a_min)/(self->cosmo->N_a-1.0);
      prefactor = 9.0/4.0*dsqr((self->cosmo->Omega_m+self->cosmo->Omega_nu_mass)/R_HUBBLE/R_HUBBLE);

      self->Pkappa = init_interTable_arr(Nzcorr, N_s, logsmin, logsmax, ds, 1.0, -3.0, err);

      for (ii=0; ii<Nzbin; ii++) {
	 for (jj=ii; jj<Nzbin; jj++) {

	    if (self->tomo==tomo_auto_only && ii!=jj) continue;
	    if (self->tomo==tomo_cross_only && ii==jj) continue;

	    table = self->Pkappa[idx_zz(ii,jj,Nzbin)]->table;
	    intpar.self = self;
	    intpar.i    = ii;
	    intpar.j    = jj;
	    for (i=0,slog=logsmin; i<N_s; i++,slog+=ds) {
	       ss = exp(slog);
	       intpar.r = ss;

#ifndef fastxi
	       /* Romberg-integration (slow) */
	       if (self->cosmo->a_min<0.7) {
		  f1 = sm2_qromberg(int_for_p_2, (void*)&intpar, self->cosmo->a_min,
				    0.7, 1.0e-6, err);
		  forwardError(*err, __LINE__, -1);
		  f2 = sm2_qrombergo(int_for_p_2, (void*)&intpar,
				     0.7, 1.0, sm2_midpntberg, 1.0e-7, err);
		  forwardError(*err, __LINE__, -1);
	       } else {
		  f1 = 0.0;
		  f2 = sm2_qrombergo(int_for_p_2, (void*)&intpar, self->cosmo->a_min,
				     1.0, sm2_midpntberg, 1.0e-7, err);
		  forwardError(*err, __LINE__, -1);
	       }
#else
	       /* Riemann summation (fast) */
	       for (a=self->cosmo->a_min,f1=0.0; a<1.0; a+=da) {
		  f1 += int_for_p_2(a, (void*)&intpar, err);
		  forwardError(*err, __LINE__, -1);
	       }
	       f1 = f1*da;
	       f2 = 0.0;
#endif

	       table[i] = log(prefactor*(f1+f2));

	       //printf("### %d %d %d %f %e\n", ii, jj, i, ss, table[i]);
	    }

	 }
      }

   }

   testErrorRetVA(self->tomo==tomo_auto_only && i_bin!=j_bin, lensing_tomoij,
		  "Cross-correlation (bins # %d,%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin, j_bin);
   testErrorRetVA(self->tomo==tomo_cross_only && i_bin==j_bin, lensing_tomoij,
		  "Cross-correlation (bin #%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin);

   slog = log(s);
   f1 = interpol_wr(self->Pkappa[idx_zz(i_bin,j_bin,Nzbin)], slog, err);
   forwardError(*err, __LINE__, -1);

   return exp(f1);
}
#undef __FUNC__

/* Used by Hankel transform functions, here and in halomodel */
#define __FUNC__ "P_projected_kappa"
double P_projected_kappa(void *self, double l, int i_bin, int j_bin, error **err)
{
   double res;
   res = Pkappa((cosmo_lens*)self, l, i_bin, j_bin, err);
   forwardError(*err, __LINE__, 0);
   return res;
}
#undef __FUNC__

int change_xi(cosmo_lens* avant, cosmo_lens* apres)
{
   return change_Pkappa(avant, apres);
}

/* ============================================================ *
 * Shear correlation functiond xi_+ (pm=+1) and xi_- (pm=-1).   *
 * ============================================================ */
#define __FUNC__ "xi"
double xi(cosmo_lens* self, int pm, double theta, int i_bin, int j_bin, error **err)
{
   double *table[2];
   double dlogtheta, logthetamin, logthetamax;
   double res;
   int Nzbin, Nzcorr, ii, jj, index;
	 
   testErrorRetVA(theta<=0, ce_negative, "Negative angular scale theta=%g", *err, __LINE__, -1, theta);

   Nzbin  = self->redshift->Nzbin;
   Nzcorr = Nzbin*(Nzbin+1)/2;
   if (self->xiP==NULL || self->xiM==NULL) {
      self->xiP = init_interTable_arr(Nzcorr, N_thetaH, 0, 1, 1, 0.0, 0.0, err);
      forwardError(*err,__LINE__,0);
      self->xiM = init_interTable_arr(Nzcorr, N_thetaH, 0, 1, 1, 0.0, 0.0, err);
      forwardError(*err,__LINE__,0);

      /* A flag could be introuced here if only auto-correlations are needed */
      for (ii=0; ii<Nzbin; ii++) {
	 for (jj=ii; jj<Nzbin; jj++) {

	    if (self->tomo==tomo_auto_only && ii!=jj) continue;
	    if (self->tomo==tomo_cross_only && ii==jj) continue;

	    index = idx_zz(ii,jj,Nzbin);
	    table[0] = self->xiP[index]->table;
	    table[1] = self->xiM[index]->table;
	    tpstat_via_hankel(self, table, &logthetamin, &logthetamax, tp_xipm, P_projected_kappa,
			ii, jj, err);
	    forwardError(*err,__LINE__,0);
	    dlogtheta     = (logthetamax-logthetamin)/((double)N_thetaH-1.0);
	    self->xiP[index]->a  = logthetamin;
	    self->xiM[index]->a  = logthetamin;
	    self->xiP[index]->b  = logthetamax;
	    self->xiM[index]->b  = logthetamax;
	    self->xiP[index]->dx = dlogtheta;
	    self->xiM[index]->dx = dlogtheta;

	 }
      }
   }

   testErrorRetVA(self->tomo==tomo_auto_only && i_bin!=j_bin, lensing_tomoij,
		  "Cross-correlation (bins # %d,%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin, j_bin);
   testErrorRetVA(self->tomo==tomo_cross_only && i_bin==j_bin, lensing_tomoij,
		  "Cross-correlation (bin #%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin);

   if (pm==1) {
      res = interpol_wr(self->xiP[idx_zz(i_bin,j_bin,Nzbin)], log(theta), err);
      forwardError(*err,__LINE__,0);
   } else if (pm==-1) {
      res = interpol_wr(self->xiM[idx_zz(i_bin,j_bin,Nzbin)], log(theta), err);
      forwardError(*err,__LINE__,0);
   } else {
      *err = addErrorVA(lensing_pm, "pm=%d not valid, has to be +1 or -1", *err, __LINE__, pm);
      res = -1.0;
      return res;
   }

   return res;
}
#undef __FUNC__

int change_gamma2(cosmo_lens* avant, cosmo_lens* apres)
{
   return change_Pkappa(avant,apres);
}

/* ============================================================ *
 * Tophat shear variance <|gamma|!2>.				*
 * ============================================================ */
#define __FUNC__ "gamma2"
double gamma2(cosmo_lens* self, double theta, int i_bin, int j_bin, error **err)
{
   double *table;
   double dlogtheta, logthetamin, logthetamax;
   double res;
   int Nzbin, Nzcorr, ii, jj, index;
   
   testErrorVA(theta<=0, ce_negative,"Negative scale theta=%g", *err, __LINE__, -1.0, theta);

   Nzbin  = self->redshift->Nzbin;
   Nzcorr = Nzbin*(Nzbin+1)/2;
   if (self->gamma==NULL) {

      self->gamma = init_interTable_arr(Nzcorr, N_thetaH, 0, 1, 1, 0.0, 0.0,err);
      forwardError(*err,__LINE__,0);

      for (ii=0; ii<Nzbin; ii++) {
	 for (jj=ii; jj<Nzbin; jj++) {

	    if (self->tomo==tomo_auto_only && ii!=jj) continue;
	    if (self->tomo==tomo_cross_only && ii==jj) continue;

	    index   = idx_zz(ii,jj,Nzbin);
	    table   = self->gamma[index]->table;
	    tpstat_via_hankel(self, &table, &logthetamin, &logthetamax, tp_gsqr, P_projected_kappa,
			      ii, jj, err);
	    forwardError(*err,__LINE__,0);
	    dlogtheta = (logthetamax-logthetamin)/((double)N_thetaH-1.0);
	    self->gamma[index]->a  = logthetamin;
	    self->gamma[index]->b  = logthetamax;
	    self->gamma[index]->dx = dlogtheta;

	 }
      }
   }

   testErrorRetVA(self->tomo==tomo_auto_only && i_bin!=j_bin, lensing_tomoij,
		  "Cross-correlation (bins # %d,%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin, j_bin);
   testErrorRetVA(self->tomo==tomo_cross_only && i_bin==j_bin, lensing_tomoij,
		  "Cross-correlation (bin #%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin);

   res = interpol_wr(self->gamma[idx_zz(i_bin,j_bin,Nzbin)], log(theta), err);
   forwardError(*err,__LINE__,0);
   return res;

}
#undef __FUNC__

int change_map2(cosmo_lens* avant, cosmo_lens* apres)
{
   return change_Pkappa(avant,apres);
}

/* ============================================================ *
 * Aperture mass variance <M_ap^2> with polynomial filter.	*
 * ============================================================ */
#define __FUNC__ "map2"
double map2(cosmo_lens* self, double theta, int i_bin, int j_bin, error **err)
{
  double res;
  res = map2_poly(self, theta, i_bin, j_bin, err);
  forwardError(*err, __LINE__, 0);
  return res;
}
#undef __FUNC__


#define __FUNC__ "map2_poly"
double map2_poly(cosmo_lens* self, double theta,  int i_bin, int j_bin, error **err)
{
   double *table;
   double dlogtheta, logthetamin, logthetamax;
   double res;
   int Nzbin, Nzcorr, ii, jj, index;


   testErrorRetVA(theta<=0, ce_negative, "Negative angular scale theta=%g", *err, __LINE__, -1, theta);

   Nzbin = self->redshift->Nzbin;
   Nzcorr = Nzbin*(Nzbin+1)/2;
   if (self->map_poly == NULL) {
      self->map_poly = init_interTable_arr(Nzcorr, N_thetaH, 0, 1, 1, 0.0, 0.0, err);
      forwardError(*err, __LINE__, 0);
      for (ii=0; ii<Nzbin; ii++) {
	 for (jj=ii; jj<Nzbin; jj++) {

	    if (self->tomo==tomo_auto_only && ii!=jj) continue;
	    if (self->tomo==tomo_cross_only && ii==jj) continue;

	    index   = idx_zz(ii,jj,Nzbin);
	    table   = self->map_poly[index]->table;
	    tpstat_via_hankel(self, &table, &logthetamin, &logthetamax, tp_map2_poly, P_projected_kappa,
			      ii, jj, err);
	    forwardError(*err, __LINE__, 0);
	    dlogtheta = (logthetamax-logthetamin)/((double)N_thetaH-1.0);
	    self->map_poly[index]->a = logthetamin;
	    self->map_poly[index]->b = logthetamax;
	    self->map_poly[index]->dx = dlogtheta;

	 }
      }
   }

   testErrorRetVA(self->tomo==tomo_auto_only && i_bin!=j_bin, lensing_tomoij,
		  "Cross-correlation (bins # %d,%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin, j_bin);
   testErrorRetVA(self->tomo==tomo_cross_only && i_bin==j_bin, lensing_tomoij,
		  "Cross-correlation (bin #%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin);

   res = interpol_wr(self->map_poly[idx_zz(i_bin,j_bin,Nzbin)], log(theta), err);
   forwardError(*err, __LINE__, 0);

   return res;
}
#undef __FUNC__

#define __FUNC__ "map2_gauss"
double map2_gauss(cosmo_lens* self, double theta, int i_bin, int j_bin, error **err)
{
   double *table;
   double dlogtheta, logthetamin, logthetamax;
   double res;
   int Nzbin, Nzcorr, ii, jj, index;

   testErrorRetVA(theta<=0, ce_negative, "Negative angular scale theta=%g", *err, __LINE__, -1, theta);

   Nzbin = self->redshift->Nzbin;
   Nzcorr = Nzbin*(Nzbin+1)/2;
   if (self->map_gauss==NULL) {
      self->map_gauss = init_interTable_arr(Nzcorr, N_thetaH, 0, 1, 1, 0.0, 0.0, err);
      forwardError(*err, __LINE__, 0);
      for (ii=0; ii<Nzbin; ii++) {
	 for (jj=ii; jj<Nzbin; jj++) {

	    if (self->tomo==tomo_auto_only && ii!=jj) continue;
	    if (self->tomo==tomo_cross_only && ii==jj) continue;

	    index   = idx_zz(ii,jj,Nzbin);
	    table   = self->map_gauss[index]->table;
	    tpstat_via_hankel(self, &table, &logthetamin, &logthetamax, tp_map2_gauss, P_projected_kappa,
			      ii, jj, err);
	    forwardError(*err, __LINE__, 0);
	    dlogtheta = (logthetamax-logthetamin)/((double)N_thetaH-1.0);
	    self->map_gauss[index]->a = logthetamin;
	    self->map_gauss[index]->b = logthetamax;
	    self->map_gauss[index]->dx = dlogtheta;
	 }
      }
   }

   testErrorRetVA(self->tomo==tomo_auto_only && i_bin!=j_bin, lensing_tomoij,
		  "Cross-correlation (bins # %d,%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin, j_bin);
   testErrorRetVA(self->tomo==tomo_cross_only && i_bin==j_bin, lensing_tomoij,
		  "Cross-correlation (bin #%d) not valid if tomo=tomo_auto_only",
		  *err, __LINE__, -1.0, i_bin);

   res = interpol_wr(self->map_gauss[idx_zz(i_bin,j_bin,Nzbin)], log(theta), err);
   forwardError(*err, __LINE__, 0);

   return res;
}
#undef __FUNC__

#define __FUNC__ "int_for_map2_slow"
double int_for_map2_slow(double logell, void *intpar, error **err)
{
   cosmo_lensANDiiid *cplusplus;
   double theta, etasqr, res, ell;
   tpstat_t tpstat;
   int i_bin, j_bin;

   ell       = exp(logell);
   cplusplus = (cosmo_lensANDiiid*)intpar;
   theta     = cplusplus->r;
   etasqr    = dsqr(theta*ell);
   tpstat    = (tpstat_t)cplusplus->t;
   i_bin     = cplusplus->i;
   j_bin     = cplusplus->j;

   res = dsqr(ell)*Pkappa(cplusplus->self, ell, i_bin, j_bin, err);
   forwardError(*err, __LINE__, 0);
   if (tpstat==tp_map2_poly) {
      /* TODO: implement Bessel function J_4 */
      *err = addError(ce_unknown, "wrong tpstat", *err, __LINE__);
      return -1;
   } else if (tpstat==tp_map2_gauss) {
      res *= dsqr(etasqr)/4.0*exp(-etasqr);
   } else {
      *err = addError(ce_unknown, "wrong tpstat", *err, __LINE__);
      return -1;
   }

   return res;
}
#undef __FUNC__

#define __FUNC__ "map2_slow"
double map2_slow(cosmo_lens *self, double theta, tpstat_t tpstat, int i_bin, int j_bin, error **err)
{
   cosmo_lensANDiiid intpar;
   double res;

   intpar.self = self;
   intpar.t    = (int)tpstat;
   intpar.r    = theta;
   intpar.i    = i_bin;
   intpar.j    = j_bin;

   res = 1.0/(2.0*pi)*sm2_qromberg(int_for_map2_slow, (void*)&intpar, log(s_min), log(s_max),
				   1.0e-6, err);
   forwardError(*err, __LINE__, 0);

   return res;
}
#undef __FUNC__

/* ============================================================ *
 * Data input, likelihood stuff (used to be in likeli.c).	*
 * ============================================================ */

/* ============================================================ *
 * (Still too quirky) initialization of lensing and n(z) data.  *
 * ============================================================ */

#define __FUNC__ "init_data_cov"
datcov* init_data_cov(char* dataname, char* covname, lensdata_t type, int isnofz, error **err)
{
   datcov *res;
   double det;
   char *data_name, *cov_name;
   char odata_name[] = "map2";
   char ocov_name[]  = "CovMap51_0.93+B.block";
   int i;
	
   if (strcmp(dataname,"")==0) {
      data_name=odata_name;
   } else {
      data_name=dataname;
   }
   if (strcmp(covname,"")==0 && isnofz!=1) {
      cov_name=ocov_name;
   } else {
      cov_name=covname;
   }

   res = malloc_err(sizeof(datcov), err);        forwardError(*err,__LINE__,NULL);
   read_data(res, data_name, err, isnofz);       forwardError(*err,__LINE__,NULL);

   if (isnofz==0 || isnofz==2) {
      read_cov(res, cov_name, err);              forwardError(*err,__LINE__,NULL);
      det = sm2_inverse(res->cov, res->n, err);  forwardError(*err,__LINE__,NULL);
      res->usecov = 1;
   } else if (isnofz==1) {
      for (i=0,det=1.0; i<res->n; i++) {
	 det *= res->var[i];
      }
      res->usecov = 0;
   } else {
      *err = addError(ce_unknown, "Unknown flag isnofz (has to be 0, 1 or 2)", *err, __LINE__);
   }

   res->type = type;

   testErrorRet(type<0||type>=Nlensdata_t, ce_unknown, "Unknown lens data type", *err, __LINE__, NULL);
   testErrorRet((isnofz==0||isnofz==2)&&type==nofz, lensing_inconsistent,
		"Inconsistent flags type and isnofz",
		*err, __LINE__, NULL);
   testErrorRet(isnofz==1&&type!=nofz, lensing_inconsistent, "Inconsistent flags type and isnofz",
		*err, __LINE__, NULL);

   return res;
}
#undef __FUNC__

/* ============================================================ *
 * Returns the index of the (i_bin j_bin)-correlation if these  *
 * are ordered lexically as					*
 * 0 0								*
 * 0 1								*
 * ...								*
 * 0 n-1							*
 * 1 1								*
 * ...								*
 * n-1 n-1							*
 * ============================================================ */
int idx_zz(int i_bin, int j_bin, int Nzbin)
{
   return i_bin*Nzbin - i_bin*(i_bin-1)/2 + j_bin - i_bin;
}

/* Index of (i i_bin j_bin)-correlation.			*/
int idx_tzz(int i, int i_bin, int j_bin, int Ntheta, int Nzbin)
{
   int Nzcorr;
   Nzcorr = Nzbin*(Nzbin+1)/2;
   return i*Nzcorr + idx_zz(i_bin, j_bin, Nzbin);
}

/* Covariance index */
int idx_tzztzz(int i, int i_bin, int j_bin, int j, int k_bin, int l_bin, int Ntheta, int Nzbin)
{
   int Nzcorr;
   Nzcorr = Nzbin*(Nzbin+1)/2;
   return idx_tzz(i, i_bin, j_bin, Ntheta, Nzbin)*Ntheta*Nzcorr
     + idx_tzz(j, k_bin, l_bin, Ntheta, Nzbin);
}

#define __FUNC__ "init_data_cov_tomo"
datcov* init_data_cov_tomo(char* dataname, char* covname, lensdata_t type, lensformat_t format, error **err)
{
   datcov *res;
   double det;
   int Nzbin;
	
   testErrorRetVA(type<0||type>=Nlensdata_t, ce_unknown, "Unknown lens data type %d", *err, __LINE__, NULL, type);

   res = malloc_err(sizeof(datcov), err);           forwardError(*err,__LINE__,NULL);
   Nzbin = 0; /* Will be set according to data file */
   printf("DATA NOT READ!!\n");
   //read_data_tomo(res, dataname, 0, err);           forwardError(*err,__LINE__,NULL);

   if (format==jan) {
      res->Ntheta = 12;
      res->Nzbin  = 6;
      res->Nzcorr = 21;
      read_cov_tomo_jan(res, covname, err);         forwardError(*err,__LINE__,NULL);
   } else if (format==benjamin) {
      read_data_tomo(res, dataname, 0, err);  forwardError(*err,__LINE__,NULL);
      read_cov_tomo_benjamin(res, covname, err);       forwardError(*err,__LINE__,NULL);
   } else {
      *err = addErrorVA(lensing_unknown, "Unknown lens format %d", *err, __LINE__, format);
      return NULL;
   }

   det = sm2_inverse(res->cov, res->n, err);        forwardError(*err,__LINE__,NULL);

   res->usecov = 1;
   res->type   = type;

   return res;
}
#undef __FUNC__

#define __FUNC__ "del_data_cov"
void del_data_cov(datcov** dc)
{
   datcov *d;
   d = *dc;

   free(d->theta);
   free(d->data);
   if (d->usecov) free(d->cov);
   else free(d->var);
   free(d);
   d = NULL;
}
#undef __FUNC__

#define __FUNC__ "read_data"
void read_data(datcov *dc, char data_name[], error **err, int isnofz)
{
   FILE *F;
   int n, res, i;
   double d[6];
   char str[1024];

   F = fopen(data_name, "r");
   if (F==NULL) {
      sprintf(str, "Cannot open data file %s. Make sure it exists and is readable", data_name);
      *err = addError(ce_file, str, *err, __LINE__);
      return;
   }

   /* Count line number */
   n = 0;
   do {
		
      res = fscanf(F, "%lf %lf %lf %lf %lf %lf\n", d, d+1, d+2, d+3, d+4, d+5);
      testErrorRet(res<2 && res!=EOF, ce_badFormat, "Number of columns <2", *err, __LINE__,);
      n++;
		
   } while (res!=EOF);
   n--;

   dc->n      = n;
   dc->Ntheta = n;
   dc->data  = malloc_err(sizeof(double)*n, err);               forwardError(*err, __LINE__,);
   dc->theta = malloc_err(sizeof(double)*n, err);               forwardError(*err, __LINE__,);
   if (isnofz==1) dc->var = malloc_err(sizeof(double)*n, err);  forwardError(*err, __LINE__,);

   /* Read file */
   rewind(F);
   for (i=0; i<n; i++) {
      fscanf(F, "%lf %lf %lf %lf %lf %lf\n", dc->theta+i, d+1, d+2, d+3, d+4, d+5);
      if (isnofz==0 || isnofz==1) dc->data[i] = d[1];
      else if (isnofz==2) dc->data[i] = d[3];
      if (isnofz==1) dc->var[i] = d[2];
   }
	 
   if (!isnofz) {
      /* [theta] from arcmin to rad */
      for (i=0; i<dc->n; i++) {
	 dc->theta[i] *= arcmin;
      }
   }

   dc->Nzbin = 1;

}
#undef __FUNC__

/* ============================================================ *
 * Reads data file into structure dc (has to be initialised     *
 * before). The file has to have the structure                  *
 *   theta[arcmin] c_00 c_01 ... c_0{n-1} c_11 ... c_{n-1}{n-1} *
 * with c_ij some (correlation) quantity corr. to bins i, j.    *
 * If Nzbin>0 on input performs a consistency check.		*
 * ============================================================ */
#define __FUNC__ "read_data_tomo"
void read_data_tomo(datcov *dc, char data_name[], int Nzbin, error **err)
{
   int n, i, i_bin, j_bin, Ncol;
   size_t Nrec, Ntheta;
   double dNcol, dNzbin;
   double *ptr;

   /* Read file to double pointer */
   Nrec = 0;
   ptr = (double*)read_any_list_count(data_name, &Nrec, "%lg", sizeof(double), &Ntheta, err);
   forwardError(*err, __LINE__,);

   /* Number of lines and columns */
   dNcol = (double)Nrec/(double)Ntheta;
   Ncol = Nrec/Ntheta;
   testErrorRet(fabs(dNcol-(double)Ncol)>epsilon, lensing_inconsistent, "Columns seem not to have equal length",
		*err, __LINE__,);

   dNzbin = 0.5*(sqrt(8.0*Ncol-7.0) - 1.0);
   testErrorRet(fabs(dNzbin-round(dNzbin))>epsilon, lensing_inconsistent, "Number of z-correlations inconsistent",
		*err, __LINE__,);
   n = (int)round(dNzbin);

   if (Nzbin>0) {
      testErrorRetVA(Nzbin!=n, lensing_inconsistent, "Number of z bins in file (%d) inconsistent with input value (%d)",
		     *err, __LINE__,, n, Nzbin);
   }
   Nzbin = n;


   /* Fill datcov structure */
   dc->Ntheta = (int)Ntheta;
   dc->Nzbin  = Nzbin;
   dc->Nzcorr = Ncol - 1;
   testErrorRet(dc->Nzcorr!=dc->Nzbin*(dc->Nzbin+1)/2, lensing_inconsistent, "Nzcorr!=Nzbin*(Nzbin+1)/2",
		*err, __LINE__,);
   dc->n      = Ntheta*dc->Nzcorr;

   dc->data  = malloc_err(sizeof(double)*dc->n, err);          forwardError(*err, __LINE__,);
   dc->theta = malloc_err(sizeof(double)*Ntheta, err);         forwardError(*err, __LINE__,);
   dc->var   = NULL;

   /* First column: angular scale */
   for (i=0; i<Ntheta; i++) {
      dc->theta[i] = ptr[i*Ncol+0]*arcmin;
   }
   /* Subsequent columns: correlation data */
   for (i=0,n=0; i<Ntheta; i++) {
      for (i_bin=0; i_bin<Nzbin; i_bin++) {
	 for (j_bin=i_bin; j_bin<Nzbin; j_bin++,n++) {
	    testErrorRetVA(n>=dc->n, math_overflow, "Index overflow (%d>=%d)",
			   *err, __LINE__,, n, dc->n);
	    dc->data[n] = ptr[i*Ncol+1+idx_zz(i_bin,j_bin,Nzbin)];
	 }
      }
   }

   /*
   for (i=0; i<Ntheta; i++) {
      for (i_bin=0; i_bin<Nzbin; i_bin++) {
	 for (j_bin=i_bin; j_bin<Nzbin; j_bin++) {
	    printf("%2d  %d %d   %f\n", i+1, i_bin+1, j_bin+1,
		   dc->data[i*dc->Nzcorr+idx_zz(i_bin, j_bin, Nzbin)]);
	 }
      }
   }
   */

}
#undef __FUNC__

/* This function is best called only from init_data_cov.    *
 * read_data_tomo has to be called before this function, or *
 * call with dc->Ntheta = dc->Nzbin = 0.		    */
#define __FUNC__ "read_cov_tomo_benjamin"
void read_cov_tomo_benjamin(datcov* dc, char cov_name[], error **err)
{
   int n, i, j, i_bin, j_bin, k_bin, l_bin, Ncol, Nzcorr, row, col, index, swap, *flag, jork;
   int i0, j0, k0, l0;
   double dNtheta, dNzcorr, dNzbin, dNcol;
   size_t Nrec, Nrow;
   double *ptr;

   /* Read file to double pointer */
   Nrec = 0;
   ptr = (double*)read_any_list_count(cov_name, &Nrec, "%lg", sizeof(double), &Nrow, err);
   forwardError(*err, __LINE__,);

   dNtheta = 0.5*(sqrt(8.0*Nrow+1.0) - 1.0);
   testErrorRetVA(fabs(dNtheta-round(dNtheta))>epsilon, lensing_inconsistent,
		"Number of angular bins inconsistent, Nrow=%d != Ntheta*(Ntheta+1/2)",
		  *err, __LINE__,, Nrow);

   if (dc->Ntheta>0) {
      testErrorRetVA((int)round(dNtheta)!=dc->Ntheta, lensing_inconsistent,
		     "Number of angular bins in file (%d) inconsistent with input value from data file (%d)",
		     *err, __LINE__,, (int)round(dNtheta), dc->Ntheta);
   } else {
      dc->Ntheta = (int)round(dNtheta);
   }

   dNcol = (double)Nrec/(double)Nrow;
   Ncol = Nrec/Nrow;
   testErrorRet(fabs(dNcol-(double)Ncol)>epsilon, lensing_inconsistent, "Columns seem not to have equal length",
		*err, __LINE__,);

   dNzcorr = 0.5*(sqrt(8.0*Ncol-15.0) - 1.0);
   Nzcorr  = (int)round(dNzcorr);
   if (dc->Nzbin>0) {
      testErrorRetVA(Nzcorr!=dc->Nzbin*(dc->Nzbin+1)/2, lensing_inconsistent,
		     "Number of z-correlations in file (%d) inconsistent with value from data file (%d=%d*(%d+1)/2)",
		     *err, __LINE__,, Nzcorr, dc->Nzbin*(dc->Nzbin+1)/2, dc->Nzbin, dc->Nzbin);
   } else {
      dNzbin = 0.5*(sqrt(8.0*dNzcorr-7.0) - 1.0);
      testErrorRet(fabs(dNzbin-round(dNzbin))>epsilon, lensing_inconsistent, "Number of z-correlations inconsistent",
		   *err, __LINE__,);
      dc->Nzbin = (int)round(dNzbin);
   }


   dc->cov = malloc_err(dc->n*dc->n*sizeof(double), err);   forwardError(*err, __LINE__,);

   /* Indicate values obtained from file, for later symmetrisation */
   flag    = calloc_err(dc->n*dc->n, sizeof(int), err);     forwardError(*err, __LINE__,);

   /* Go through covariance file, calculate for each entry matrix index *
    * jork: from Patrick Simon's black magic code			*/
   for (i=0,row=0; i<dc->Ntheta; i++) {
      for (j=i; j<dc->Ntheta; j++,row++) {

	 for (i_bin=0,col=0; i_bin<dc->Nzbin; i_bin++) {
	    for (j_bin=i_bin; j_bin<dc->Nzbin; j_bin++) {
	       jork = 0;
	       for (k_bin=i_bin; k_bin<dc->Nzbin; k_bin++,jork=1) {
		  for (l_bin=(jork==0?j_bin:k_bin); l_bin<dc->Nzbin; l_bin++) {

		     index = idx_tzztzz(i, i_bin, j_bin, j, k_bin, l_bin, dc->Ntheta, dc->Nzbin);

		     testErrorRetVA(index>=dc->n*dc->n, math_overflow, "Index overflow (%d>=%d)",
				  *err, __LINE__,, index, dc->n*dc->n);
		     dc->cov[index] = ptr[row*Ncol+2+col];
		     flag[index]    = 1;
		     //printf("READ %d %d   %d %d  %d %d   %g\n", i+1, j+1, i_bin+1, j_bin+1, k_bin+1, l_bin+1, 
		     //dc->cov[index]);

		     col++;
		  }
	       }
	    }
	 }
      }
   }

   /* Fill missing (symmetrical) elements: Go through matrix, swap indices to get *
    * entry which  was read from file. */
   for (i=0; i<dc->Ntheta; i++) {
      for (j=0; j<dc->Ntheta; j++) {

	 for (i_bin=0,col=0; i_bin<dc->Nzbin; i_bin++) {
	    for (j_bin=i_bin; j_bin<dc->Nzbin; j_bin++) {

	       for (k_bin=0; k_bin<dc->Nzbin; k_bin++) {
		  for (l_bin=k_bin; l_bin<dc->Nzbin; l_bin++) {

		     index = idx_tzztzz(i, i_bin, j_bin, j, k_bin, l_bin, dc->Ntheta, dc->Nzbin);

		     if (flag[index]==0) {
			flag[index] = 1;

			/* Check whether pairs (ij),(kl) have to be swapped */
			swap = 0;
			if (MIN(i_bin,j_bin)>MIN(k_bin,l_bin)) swap = 1;
			if (MIN(i_bin,j_bin)>=MIN(k_bin,l_bin) && MAX(i_bin,j_bin)>MAX(k_bin,l_bin)) swap = 1;

			if (swap==0) {
			   i0 = MIN(i_bin,j_bin);
			   j0 = MAX(i_bin,j_bin);
			   k0 = MIN(k_bin,l_bin);
			   l0 = MAX(k_bin,l_bin);
			} else {
			   i0 = MIN(k_bin,l_bin);
			   j0 = MAX(k_bin,l_bin);
			   k0 = MIN(i_bin,j_bin);
			   l0 = MAX(i_bin,j_bin);
			}
			n = idx_tzztzz(MIN(i,j),i0,j0,MAX(i,j),k0,l0,dc->Ntheta,dc->Nzbin);

			//printf("%s %3d %3d    ", swap==0?"k":"s", index, n);

			dc->cov[index] = dc->cov[n];

		     } else {
			/* Already calculated, do nothing. */
			//printf("o %3d %3d    ", index, index);
		     }
		     //printf("%d %d   %d %d  %d %d   %g\n", i+1, j+1, i_bin+1, j_bin+1, k_bin+1, l_bin+1, dc->cov[index]);

		  }
	       }
	    }
	 }
      }
   }

   free(flag);

}
#undef __FUNC__

#define __FUNC__ "read_cov_tomo_jan"
void read_cov_tomo_jan(datcov* dc, char cov_name[], error **err)
{
   int i, j, i_bin, j_bin, k_bin, l_bin, index_cov, index_file;
   size_t Nrec, Nrow;
   double *ptr;

   testErrorRet(dc->Ntheta<=0, lensing_initialised, "Data file has to be read before covariance, to set number of bins", *err, __LINE__,);
   testErrorRet(dc->Nzcorr<=0, lensing_initialised, "Data file has to be read before covariance, to set number of bins", *err, __LINE__,);
   testErrorRet(dc->Nzbin<=0, lensing_initialised, "Data file has to be read before covariance, to set number of bins", *err, __LINE__,);
   testErrorRetVA(dc->type!=xipm, lensing_inconsistent, "Type has to be xipm and not (%d)%s\n", *err, __LINE__,, dc->type, slensdata_t(dc->type));

   /* Read file to double pointer */
   Nrec = 0;
   ptr = (double*)read_any_list_count(cov_name, &Nrec, "%lg", sizeof(double), &Nrow, err);
   forwardError(*err, __LINE__,);

   printf("Nrec = %zu Nrow = %zu\n", Nrec, Nrow);
   testErrorRetVA(Nrec!=Nrow*Nrow, lensing_inconsistent, "Covariance matrix is not square, Nrow=%d, Nrec=%d\n",
		  *err, __LINE__,, Nrec, Nrow);

   for (i_bin=0,index_file=0; i_bin<dc->Nzbin; i_bin++) {
      for (j_bin=i_bin; j_bin<dc->Nzbin; j_bin++) {
	 for (i=0; i<dc->Ntheta; i++) {

	    for (k_bin=0; k_bin<dc->Nzbin; k_bin++) {
	       for (l_bin=k_bin; l_bin<dc->Nzbin; l_bin++) {
		  for (j=0;j<dc->Ntheta;j++,index_file++) {

		     index_cov = idx_tzztzz(i, i_bin, j_bin, j, k_bin, l_bin, dc->Ntheta, dc->Nzbin);
		     dc->cov[index_cov] = ptr[index_file];
		  }
	       }
	    }
	 }
      }
   }

}
#undef __FUNC__

#define __FUNC__ "read_cov"
void read_cov(datcov* dc, char cov_name[], error **err)
{
   FILE *F;
   int n, i;
	
   F       = fopen_err(cov_name, "r", err);                 forwardError(*err, __LINE__,);
   dc->cov = malloc_err(sizeof(double)*dc->n*dc->n, err);   forwardError(*err, __LINE__,);
	
   for (i=0; i<dc->n*dc->n; i++) {
      n=fscanf(F,"%lf",&(dc->cov[i]));
      if (n!=1 || n==EOF) {
	 *err = addError(ce_badFormat,"bad format",*err,__LINE__);
	 return;
      }
   }
   fclose(F);
}
#undef __FUNC__

/* Really only reads a covariance (in column format) for xi+ */
#define __FUNC__ "read_cov_col"
void read_cov_col(datcov *dc, char cov_name[], error **err)
{
   FILE *F;
   int i, j, res, nn;
   unsigned int ncomment;
   double dummy[5];

   nn        = numberoflines_comments(cov_name, &ncomment, err); forwardError(*err, __LINE__,);
   dc->n     = (int)sqrt(nn);
   F         = fopen_err(cov_name, "r", err);                 forwardError(*err, __LINE__,);
   dc->cov   = malloc_err(sizeof(double)*dc->n*dc->n, err);   forwardError(*err, __LINE__,);
   dc->theta = malloc_err(sizeof(double)*dc->n, err);         forwardError(*err, __LINE__,);

   for (i=0; i<dc->n; i++) {
      for (j=0; j<dc->n; j++) {
	 res = fscanf(F, "%lf %lf %lf %lf %lf\n", dummy, dummy+1, dummy+2, dummy+3, dummy+4);
	 testErrorRet(res==EOF, io_eof, "Premature eof", *err, __LINE__,);
	 if (i==0) {
	    dc->theta[j] = dummy[1];
	 }
	 dc->cov[i*dc->n+j] = dummy[2];
      }
   }

}
#undef __FUNC__

/* ============================================================ *
 * Lensing signal, shear second-order statistics for angular    *
 * scale theta [rad].						*
 * ============================================================ */

#define __FUNC__ "lensing_signal"
double lensing_signal(cosmo_lens *model, double theta, int i_bin, int j_bin, lensdata_t type, error **err)
{
   double res;

   switch (type) {
      case xip : case xiE :
	 res = xi(model, +1, theta, i_bin, j_bin, err);
	 break;
      case xim :
	 res = xi(model, -1, theta, i_bin, j_bin, err);
	 break;
      case map2poly :
	 res = map2_poly(model, theta, i_bin, j_bin, err);
	 break;
      case map2gauss :
	 res = map2_gauss(model, theta, i_bin, j_bin, err);
	 break;
      case gsqr :
	 res = gamma2(model, theta, i_bin, j_bin, err);
	 break;
      case nofz :
      default :
	 *err = addErrorVA(ce_unknown, "Unknown lensdata type %d", *err, __LINE__, type);
	 return 0.0;
   }

   forwardError(*err, __LINE__, 0);
   return res;
}
#undef __FUNC__

#define __FUNC__ "chi2_lensing"
double chi2_lensing(cosmo_lens* csm, datcov* dc, error **err)
{
   double *data_minus_model, model;
   int i, j, in, in1, in2, i_bin, j_bin, k_bin, l_bin, Nzbin;
   double res;

   /* Unphysically high baryon fraction, caused infinite chi^2 values. */
   testErrorRet(csm->cosmo->Omega_b/csm->cosmo->Omega_m>BARYON_FRAC, lensing_baryon_fraction,
		"Baryon fraction unphysically high", *err, __LINE__, 0.0);

   Nzbin = csm->redshift->Nzbin;
   testErrorRetVA(Nzbin!=dc->Nzbin, redshift_Nzbin,
		  "Number of redshift bins for model (%d) inconsistent with data (%d)",
		  *err, __LINE__, 0, Nzbin, dc->Nzbin);

   data_minus_model = malloc_err(sizeof(double)*dc->n, err);
   forwardError(*err, __LINE__, 0);

   model = lensing_signal(csm, 1.0*arcmin, 0, 0, dc->type, err);
   forwardError(*err, __LINE__, 0);

   for (j=0; j<dc->Ntheta; j++) {
      for (i_bin=0; i_bin<Nzbin; i_bin++) {
	 for (j_bin=i_bin; j_bin<Nzbin; j_bin++) {
	    model = lensing_signal(csm, dc->theta[j], i_bin, j_bin, dc->type, err);
	    forwardError(*err, __LINE__, -1);

	    in = idx_tzz(j, i_bin, j_bin, dc->Ntheta, Nzbin);
	    testErrorRet(in>=dc->n, math_overflow, "i>in overflow", *err, __LINE__, 0);
	    data_minus_model[in] = dc->data[in] - model;

	    //printf("%3d %.3f % .5e % .5e\n", j, dc->theta[j]/arcmin, dc->data[in], model);
	 }
      }
   }

   res = 0.0;

   if (dc->usecov) {

      /* This could be done simpler, with just two loops up to dc->n.   *
       * Even better: use mvdens routines.				*/
      for (i=0;i<dc->Ntheta;i++) {
	 for (i_bin=0; i_bin<Nzbin; i_bin++) {
	    for (j_bin=i_bin; j_bin<Nzbin; j_bin++) {
	       in1 = idx_tzz(i,i_bin,j_bin,dc->Ntheta,dc->Nzbin);
	       //in = i*dc->Nzcorr + idxt(i_bin,j_bin,Nzbin);

	       for (j=0;j<dc->Ntheta;j++) {
		  for (k_bin=0; k_bin<Nzbin; k_bin++) {
		     for (l_bin=k_bin; l_bin<Nzbin; l_bin++) {

			in = idx_tzztzz(i, i_bin, j_bin, j, k_bin, l_bin, dc->Ntheta, dc->Nzbin);
			testErrorRetVA(in>=dc->n*dc->n, math_overflow, "overflow (%d>=%d", *err, __LINE__, 0,
				       in, dc->n*dc->n);

			in2 = idx_tzz(j,k_bin,l_bin,dc->Ntheta,dc->Nzbin);

			res += data_minus_model[in1]*dc->cov[in]*data_minus_model[in2];

			//printf("%3d %3d % .5e % .5e % .5e\n",
			//   i, j, data_minus_model[in1], dc->cov[in], data_minus_model[in2]);
		     }
		  }
	       }

	    }
	 }
      }

   } else {

      /* TODO!! Needed anyway? */
      for (i=0;i<dc->n;i++) {
	 res += dsqr(data_minus_model[i]/dc->var[i]);
      }

   }

   testErrorRet(!isfinite(res), ce_infnan, "inf or nan in logl", *err, __LINE__, -1);
   free(data_minus_model);

   //dump_param_lens(csm, stderr, err);
   //forwardError(*err, __LINE__, 0.0);

   /* det C ... */
   res = -0.5*res;

   //printf("chi2_lensing: Returning %g\n", res);

   return res;
}
#undef __FUNC__
