/* ============================================================ *
 * nofz.c						        *
 * Martin Kilbinger 2009					*
 * ============================================================ */

#include "nofz.h"

/* ============================================================ *
 * Initialised redshift without filling in data.		*
 * ============================================================ */
#define __FUNC__ "init_redshift_empty"
redshift_t *init_redshift_empty(int Nzbin, int Nnz_max, error **err)
{
   redshift_t *res;

   res = malloc_err(sizeof(redshift_t), err);   forwardError(*err, __LINE__, NULL);
 
   res->Nzbin     = Nzbin;
   res->Nnz       = calloc_err(Nzbin, sizeof(int), err);                forwardError(*err, __LINE__, NULL);
   res->Nnz_max   = Nnz_max;
   res->nofz      = calloc_err(Nzbin, sizeof(nofz_t), err);             forwardError(*err, __LINE__, NULL);
   res->par_nz    = calloc_err(Nzbin*Nnz_max, sizeof(double), err);     forwardError(*err, __LINE__, NULL);
   res->prob_norm = calloc_err(Nzbin, sizeof(double), err);             forwardError(*err, __LINE__, NULL);
   forwardError(*err, __LINE__, NULL);

   return res;
}
#undef __FUNC__

/* ============================================================= *
 * Initialises redshift, copies content of the Nzbin-dimensional *
 * arrays Nnz, nofz and par_nz.					 *
 * ============================================================= */
#define __FUNC__ "init_redshift"
redshift_t *init_redshift(int Nzbin, const int *Nnz, const nofz_t *nofz, const double *par_nz, error **err)
{
   redshift_t *res;
   int n;

   /* The upper limit 1000 is arbitrary, mainly a test whether Nzbin is initialised properly. *
    * We don't claim that the code works with 999 redshift bins.			      */
   testErrorRetVA(Nzbin<=0 || Nzbin>1000, redshift_Nzbin, "Number of redshift bins (Nzbin=%d) not valid",
		  *err, __LINE__, NULL, Nzbin);

   for (n=0; n<Nzbin; n++) {
      testErrorRet(Nnz[n]<2, redshift_nnz, "Number of redshift parameters must be >= 2",
		   *err, __LINE__, 0);
   }

   res = malloc_err(sizeof(redshift_t), err);   forwardError(*err, __LINE__, NULL);

   res->Nzbin = Nzbin;

   res->Nnz = malloc_err(Nzbin*sizeof(int), err); forwardError(*err, __LINE__, NULL);
   memcpy(res->Nnz, Nnz, Nzbin*sizeof(int));

   for (n=0,res->Nnz_max=-1; n<Nzbin; n++) {
      if (res->Nnz[n]>res->Nnz_max) res->Nnz_max = res->Nnz[n];
   }
   
   res->nofz = malloc_err(Nzbin*sizeof(nofz), err); forwardError(*err, __LINE__, NULL);
   memcpy(res->nofz, nofz, Nzbin*sizeof(nofz));

   res->par_nz = malloc_err(Nzbin*res->Nnz_max*sizeof(double), err);
   forwardError(*err, __LINE__, NULL);
   memcpy(res->par_nz, par_nz, Nzbin*res->Nnz_max*sizeof(double));

   res->prob_norm = calloc_err(Nzbin, sizeof(double), err);
   forwardError(*err, __LINE__, NULL);

   return res;
}
#undef __FUNC__

#define __FUNC__ "copy_redshift"
redshift_t *copy_redshift(const redshift_t *REDSHIFT, error **err)
{
   redshift_t *res;

   res = init_redshift(REDSHIFT->Nzbin, REDSHIFT->Nnz, REDSHIFT->nofz, REDSHIFT->par_nz, err);
   forwardError(*err, __LINE__, NULL);

   testErrorRetVA(REDSHIFT->Nnz_max!=res->Nnz_max, redshift_nnz, "Inconsistent Nz_max (%d!=%d)",
		  *err, __LINE__, NULL, REDSHIFT->Nnz_max, res->Nnz_max);

   return res;
}
#undef __FUNC__

#undef __FUNC__

#define __FUNC__ "updateFrom_redshift"
void updateFrom_redshift(const redshift_t *avant, redshift_t *apres, error **err)
{
   if (change_redshift(avant, apres)) {
      free(apres->prob_norm);
      apres->prob_norm = NULL;
   }
}
#undef __FUNC__

void free_redshift(redshift_t **self)
{
   redshift_t *s;

   s = *self;
   free(s->Nnz);          s->Nnz = NULL;
   free(s->nofz);         s->nofz = NULL;
   free(s->par_nz);       s->par_nz = NULL;
   free(s->prob_norm);    s->prob_norm = NULL;
   free(s);
   s = NULL;
}

/* Useful if redshift changes */
#define __FUNC__ "free_and_reinit_redshift"
void free_and_reinit_redshift(redshift_t **self, int Nzbin, int Nnz_max, error **err)
{
   if (*self!=NULL) free_redshift(self);
   *self = init_redshift_empty(Nzbin, Nnz_max, err);
   forwardError(*err, __LINE__,);
}
#undef __FUNC__

/* ============================================================ *
 * The mother of all functions to read redshift stuff from      *
 * files. Also used by wrappers/nz.c:nz_init.			*
 * Reads from a config-type file:				*
 *    Nzbin	[integer]					*
 *    snzmode	[string]					*
 * Next, depending on nzmode, does				*
 * - nothing (nz_fit_hist)					*
 * - reads redshift from same file (nz_read_in_situ)            *
 * - reads							*
 *      nzfile     [Nzbin strings]				*
 *   and reads redshift from files (nz_read_from_files)         *
 * ============================================================ */
#define __FUNC__ "read_redshift_info"
void read_redshift_info(redshift_t **self, FILE *F, error **err)
{
   config_element c;
   struct {
     int Nzbin;
     char snzmode[128], **nzfile;
     nzmode_t nzmode;
   } tmp;
   int j;
   char line[128], s[128];

   CONFIG_READ(&tmp, Nzbin, i, F, c, err);
   CONFIG_READ_S(&tmp, snzmode, s, F, c, err);
   STRING2ENUM(&tmp, nzmode, snzmode, nzmode_t, snzmode_t, Nnzmode_t, err);

   switch (tmp.nzmode) {

      case nz_fit_hist :
	 *self = NULL;
	 return;           /* Used only in wrappers:nz.c */

      case nz_read_in_situ :
	 read_parameters_redshift(self, F, tmp.Nzbin, err);
	 forwardError(*err, __LINE__,);
	 break;

      case nz_read_from_files :
	 tmp.nzfile = malloc_err(sizeof(char*)*tmp.Nzbin, err);
	 forwardError(*err, __LINE__,);
	 for (j=0; j<tmp.Nzbin; j++) {
	    tmp.nzfile[j] = malloc_err(sizeof(char)*128*tmp.Nzbin, err);
	 }
	 CONFIG_READ_S_ARR(&tmp, nzfile, s, j, tmp.Nzbin, F, c, err);

	 *self = init_redshift_from_files((const char**)(tmp.nzfile), tmp.Nzbin, err);
	 forwardError(*err, __LINE__,);
	 break;
 
     default :
	*err = addErrorVA(redshift_unknown, "Unknown nzmode type %d\n",
			  *err, __LINE__, tmp.nzmode);
	 return;
   }
}
#undef __FUNC__

/* Reads redshift parameters from a config-type file with format:
 * Nnz_max      [integer]
 * Nnz          [Nzbin integers] 
 * nofz         [Nzbin integers]
 * par_nz       [Nzbin*Nnzmax doubles]
 */
#define __FUNC__ "read_parameters_redshift"
void read_parameters_redshift(redshift_t **self, FILE *F, int Nzbin, error **err)
{
   redshift_t tmp, *tmp2;
   config_element c;
   int n;
   char s[500];

   CONFIG_READ((&tmp), Nnz_max, i, F, c, err);

   tmp2 = init_redshift_empty(Nzbin, tmp.Nnz_max, err);
   forwardError(*err, __LINE__,);

   CONFIG_READ_ARR(tmp2, Nnz, i, n, Nzbin, s, F, c, err);
   CONFIG_READ_ARR(tmp2, nofz, i, n, Nzbin, s, F, c, err);
   CONFIG_READ_ARR(tmp2, par_nz, d, n, Nzbin*tmp.Nnz_max, s, F, c, err);
   //dump_redshift_nostruct(Nzbin, tmp2->Nnz, tmp2->nofz, tmp2->par_nz, tmp.Nnz_max, stdout);

   *self = init_redshift(tmp2->Nzbin, tmp2->Nnz, tmp2->nofz, tmp2->par_nz, err);
   forwardError(*err, __LINE__,);

   free_redshift(&tmp2);
}
#undef __FUNC__

/* Reads parameters of slice n_bin from file name. Used by init_redshift_from_files. */
#define __FUNC__ "read_redshift_slice"
void read_redshift_slice(redshift_t *self, int n_bin, const char *name, error **err)
{
   nofz_t nofz;
   int i;
   size_t Nrec, Nlines;
   double *ptr;

   get_nofz_t_file(name, &nofz, err);
   forwardError(*err, __LINE__,);

   self->nofz[n_bin] = nofz;

   if (nofz==hist) {
      ptr = read_par_nz_hist(name, self->Nnz+n_bin, err);
      forwardError(*err, __LINE__,);
      memcpy(self->par_nz + n_bin*self->Nnz_max, ptr, self->Nnz[n_bin]*sizeof(double));
   } else {
      Nrec = 0;
      ptr = (double*)read_any_list_count(name, &Nrec, "%lg", sizeof(double), &Nlines, err);
      testErrorRetVA(Nrec!=Nlines, redshift_nnz,
		     "Number of records (%d) not equal to number of lines (%d) in n(z) file '%s'",
		     *err, __LINE__,, Nrec, Nlines, name);
      self->Nnz[n_bin] = Nlines;
      for (i=0; i<Nrec; i++) {
	 self->par_nz[n_bin*self->Nnz_max+i] = ptr[i];
      }
   }
   free(ptr);
}
#undef __FUNC__

#define __FUNC__ "dump_redshift"
void dump_redshift(redshift_t *self, FILE *F, error **err)
{
   int n_bin;

   dump_redshift_nostruct(self->Nzbin, self->Nnz, self->nofz, self->par_nz, self->Nnz_max, F);
   fprintf(F, "# Mean redshift =");
   for (n_bin=0; n_bin<self->Nzbin; n_bin++) {
      fprintf(F, " %.3f", zmean(self, n_bin, err));
      forwardError(*err, __LINE__,);
   }
   fprintf(F, "\n");
}
#undef __FUNC__

void dump_redshift_nostruct(int Nzbin, int *Nnz, nofz_t *nofz, double *par_nz, int Nnz_max, FILE *F)
{
   int n, m;

   //fprintf(stderr, "### %p %p %p\n", Nnz, nofz, par_nz);

   /* Number of redshift bins */
   fprintf(F, "# Nzbin = %d\n#", Nzbin);

   /* Number of redshift parameters */
   for (n=0; n<Nzbin; n++) fprintf(F, " nofz%d", n);
   fprintf(F, "\n#");
   for (n=0; n<Nzbin; n++) fprintf(F, "     %d", nofz[n]);
   fprintf(F, "\n# ");

   /* Redshift distribution type */
   for (n=0; n<Nzbin; n++) fprintf(F, " Nnz%d ", n);
   fprintf(F, "\n#");
   for (n=0; n<Nzbin; n++) fprintf(F, "     %d", Nnz[n]);
   fprintf(F, "\n#  ");

   /* Redshift parameters */
   for (n=0; n<Nzbin; n++) {
      for (m=0; m<Nnz[n]; m++) {
	 fprintf(F, "pnz%d   ", m);
      }
      fprintf(F, "   ");
   }
   fprintf(F, "\n# ");
   for (n=0; n<Nzbin; n++) {
      for (m=0; m<Nnz[n]; m++) {
	 fprintf(F, "%.3f  ", par_nz[n*Nnz_max+m]);
      }
      fprintf(F, "   ");
   }
   fprintf(F, "\n");
}


/* Used in lens_likeli and nz_likeli, to copy parameter values into redshift structure */
#define __FUNC__ "fill_redshift_slice"
void fill_redshift_slice(redshift_t *self, int n_bin, nofz_t nofz, error **err, ...)
{
   va_list args;
   int Nnz_max;

   va_start(args, err);

   self->nofz[n_bin] = nofz;
   Nnz_max           = self->Nnz_max;

   switch (nofz) {

      case ymmk :
	 self->Nnz[n_bin] = 5;
	 self->par_nz[n_bin*Nnz_max+0] = va_arg(args, double);   /* z_min */
	 self->par_nz[n_bin*Nnz_max+1] = va_arg(args, double);   /* z_max */
	 self->par_nz[n_bin*Nnz_max+2] = va_arg(args, double);   /* a_ymmk */
	 self->par_nz[n_bin*Nnz_max+3] = va_arg(args, double);   /* b_ymmk */
	 self->par_nz[n_bin*Nnz_max+4] = va_arg(args, double);   /* c_ymmk */
	 break;

      default :
	 *err = addErrorVA(redshift_unknown, "Unknown redshift type %d\n", *err, __LINE__, nofz);
	 return;
   }
}
#undef __FUNC__

/* ============================================================ *
 * Reads a redshift histogram from the file "name" with columns	*
 *    z_i(left bin corner)   n_i				*
 * The last row contains					*
 *    z_max x							*
 * where x is not used and can be any number.			*
 * Returns vector with redshift parameters and sets Nnz         *
 * accordingly.							*
 * ============================================================ */
#define __FUNC__ "read_par_nz_hist"
double *read_par_nz_hist(const char *name, int *Nnz, error **err)
{
   FILE *F;
   char c, cpre;
   int i, n, nread;
   double *par_nz, dummy;
   char str[1024], dummy2[128];

   /* Number of lines */
   F = fopen(name, "r");
   if (F==NULL) {
      sprintf(str, "Could not open file %s", name);
      *err = addError(io_file, str, *err, __LINE__);
      return NULL;
   }
   cpre = '\n';
   n = 0;
   do {
      c = fgetc(F);
      if (c=='#' && cpre=='\n') n--;
      if (c=='\n') n++;
      cpre = c;
   } while (c!=EOF);
   testErrorRet(n==0, io_file, "File corrupt: no valid line found in n(z) file", *err, __LINE__, NULL);
   n--;   /* Last line contains z_n 0 */

   *Nnz = 2*n+1;
   par_nz = malloc_err(sizeof(double)*(*Nnz), err);  forwardError(*err, __LINE__, NULL);

   /* Read parameters */
   /* TODO: Replace following by read_any_list */
   rewind(F);

   /* Header line */
   nread = fscanf(F, "%s %s\n", dummy2, str);
   testErrorRetVA(strcmp(dummy2, "#")!=0, io_file,
		  "Wrong format of redshift info file '%s': First line has to be '#   snofz_t'",
		  *err, __LINE__, NULL, name);
   testErrorRet(nread!=2, io_file, "Wrong format of redshift info file: First line has to be '#   snofz'",
		*err, __LINE__, NULL);

   nread = fscanf(F, "%lg %lg\n", par_nz, par_nz+n+1);           /* z_0  n_0 */
   testErrorRetVA(nread!=2, io_file, "Error while reading n(z) file ('%s'): two doubles expected (first line)",
		  *err, __LINE__, NULL, name);

   for (i=1; i<n; i++) {
      nread = fscanf(F, "%lg %lg\n", par_nz+i+1, par_nz+n+i+1);  /* z_i  n_i */
      testErrorRetVA(nread!=2, io_file, "Error while reading n(z) file ('%s'): two doubles expected",
		     *err, __LINE__, NULL, name);
   }

   nread= fscanf(F, "%lg %lg\n", par_nz+1, &dummy);            /* z_{n-1}=z_max  0 */
   testErrorRetVA(nread!=2, io_file, "Error while reading n(z) file ('%s'): two doubles expected (last line)",
		  *err, __LINE__, NULL, name);

   return par_nz;
}
#undef __FUNC__

/* Determines the redshift type of file 'name' from its header  *
 * # snofz							*/
#define __FUNC__ "get_nofz_t_file"
void get_nofz_t_file(const char *name, nofz_t *nofz, error **err)
{
   FILE *F;
   int n, j;
   struct { char snofz[128]; nofz_t nofz; } tmp;
   char line[128], dummy[128];

   /* Read first line with nofz type */
   F = fopen_err(name, "r", err);
   forwardError(*err, __LINE__,);
   n = fscanf(F, "%s %s\n", dummy, tmp.snofz);
   //printf("n=%d dummy=(%s) nofz=(%s)\n", n, dummy, tmp.snofz);
   testErrorRetVA(strcmp(dummy, "#")!=0, io_file,
		  "Wrong format of redshift info file '%s': First line has to be '#   snofz_t'",
		  *err, __LINE__,, name);
   testErrorRetVA(n!=2, io_file,
		  "Wrong format of redshift info file '%s': First line has to be '#   snofz_t'",
		  *err, __LINE__,, name);
   fclose(F);
   STRING2ENUM(&tmp, nofz, snofz, nofz_t, snofz_t, Nnofz_t, err);
   *nofz = tmp.nofz;
}
#undef __FUNC__

/* Sets on output number of parameters Nnz corresponding to file 'name' */
#define __FUNC__ "Nnz_from_file"
void Nnz_from_file(const char *name, int *Nnz, error **err)
{
   size_t Nrec, Nlines;
   nofz_t nofz;

   get_nofz_t_file(name, &nofz, err);
   forwardError(*err, __LINE__,);

   /* Read complete file for record count */
   Nrec = 0;
   read_any_list_count(name, &Nrec, "%lg", sizeof(double), &Nlines, err);
   forwardError(*err, __LINE__,);

   switch (nofz) {
      case ludo : case ymmk : case ymmk0const :
	 *Nnz = Nrec + 2;                   /* Params + zmin + zmax */
	 break;
      case hist :
	 *Nnz = Nrec - 1;                   /* All entries - last (=0) */
	 break;
      default :
	 *err = addErrorVA(redshift_unknown, "Unknown redshift type %d\n", *err, __LINE__, nofz);
   }
}
#undef __FUNC__

/* Reads n(z) file(s) of any nofz_t type and returns the corresponding redshift structure, one or more bins */
#define __FUNC__ "init_redshift_from_files"
redshift_t *init_redshift_from_files(const char **name, int Nzbin, error **err)
{
   redshift_t *res;
   int n, this_Nnz, Nnz_max;

   /* Get maximum number of redshift parameters */
   for (n=0,Nnz_max=-1; n<Nzbin; n++) {
      Nnz_from_file(name[n], &this_Nnz, err);
      forwardError(*err, __LINE__, NULL);
      if (Nnz_max<this_Nnz) Nnz_max = this_Nnz;
   }

   res = init_redshift_empty(Nzbin, Nnz_max, err);
   forwardError(*err, __LINE__, NULL);
   for (n=0; n<Nzbin; n++) {
      read_redshift_slice(res, n, name[n], err);
      forwardError(*err, __LINE__, NULL);
   }

   return res;
}
#undef __FUNC__

#define __FUNC__ "init_redshift_from_histogram_file"
redshift_t *init_redshift_from_histogram_file(const char *name, error **err)
{
   redshift_t *self;

   self = init_redshift_from_files(&name, 1, err);
   forwardError(*err, __LINE__, NULL);
   return self;

   /*
   Nnz_from_file(name, &Nnz, err);
   forwardError(*err, __LINE__, NULL);

   self = init_redshift_empty(1, Nnz, err);
   forwardError(*err, __LINE__, NULL);

   read_redshift_slice(self, 1, name, err);
   forwardError(*err, __LINE__, NULL);

   return self;
   */
}
#undef __FUNC__

/* Returns z_min for redshift bin #n_bin */
#define __FUNC__ "get_zmin"
double get_zmin(const redshift_t *redshift, int n_bin)
{
   double zmin;
   zmin = redshift->par_nz[n_bin*redshift->Nnz_max + 0];
   return zmin;
}
#undef __FUNC__

/* Returns z_max for redshift bin #n_bin */
#define __FUNC__ "get_zmax"
double get_zmax(const redshift_t *redshift, int n_bin)
{
   return redshift->par_nz[n_bin*redshift->Nnz_max + 1];
}
#undef __FUNC__

#define __FUNC__ "get_amin"
double get_amin(const redshift_t *redshift, error **err)
{
   int n;
   double amin, amintmp;

   for (n=0,amin=1.0; n<redshift->Nzbin; n++) {
      amintmp = 1.0/(1.0 + get_zmax(redshift, n));
      if (amintmp<amin) amin = amintmp;
   }
   testErrorRet(amin>=1.0, redshift_zmax, "zmax not valid (has to be different from zero)",
		*err, __LINE__, 0.0);

   return amin;
}
#undef __FUNC__

#define __FUNC__ "prob_unorm"
double prob_unnorm(double z, void *intpar, sm2_error **err)
{
   const double trunc = 100.0;    /* truncation for exp(-trunc) and smaller */
   double x, p, res, a, b, c, z0;
   redshiftANDint* rANDi;
   redshift_t *self;
   int n, m, n_bin;
   
   rANDi = (redshiftANDint*)intpar;
   self  = rANDi->self;
   n_bin = rANDi->i;

   switch (self->nofz[n_bin]) {
      case ludo :
	 testErrorRet(self->Nnz[n_bin]!=5, redshift_nnz,
		      "Wrong number of redshift parameters, should be 5 for nofz=ludo",
		      *err, __LINE__, 0);
	 a  = self->par_nz[n_bin*self->Nnz_max+2];
	 b  = self->par_nz[n_bin*self->Nnz_max+3];
	 z0 = self->par_nz[n_bin*self->Nnz_max+4];

	 testErrorRet(b<epsilon, redshift_betap, "beta_p<0", *err, __LINE__, 0);

	 x = z/z0;
	 p = pow(x, b);
	 if (p>trunc) res = 0.0;
	 res = pow(x, a)*exp(-p);
	 break;

      case jonben :
	 testErrorRet(self->Nnz[n_bin]!=5, redshift_nnz,
		      "Wrong number of redshift parameters, should be 5 for nofz=jonben",
		      *err, __LINE__, 0);
	 a = self->par_nz[n_bin*self->Nnz_max+2];
	 b = self->par_nz[n_bin*self->Nnz_max+3];
	 c = self->par_nz[n_bin*self->Nnz_max+4];
	 res = pow(z, a)/(pow(z, b) + c);
	 break;

      case ymmk :
	 testErrorRet(self->Nnz[n_bin]!=5, redshift_nnz,
		      "Wrong number of redshift parameters, should be 5 for nofz=ymmk",
		      *err, __LINE__, 0);
	 a = self->par_nz[n_bin*self->Nnz_max+2];
	 b = self->par_nz[n_bin*self->Nnz_max+3];
	 c = self->par_nz[n_bin*self->Nnz_max+4];
	 res = (pow(z, a) + pow(z, a*b))/(pow(z, b) + c);
	 break;

      case ymmk0const :
	 testErrorRet(self->Nnz[n_bin]!=7, redshift_nnz, 
		      "Wrong number of redshift parameters, should be 7 for nofz=ymmk0const",
		      *err, __LINE__, 0);
	 a = self->par_nz[n_bin*self->Nnz_max+2];
	 b = self->par_nz[n_bin*self->Nnz_max+3];
	 c = self->par_nz[n_bin*self->Nnz_max+4];
	 res = (pow(z, a) + pow(z, a*b))/(pow(z, b) + c);

	 testErrorRet(self->par_nz[n_bin*self->Nnz_max+5]<0, redshift_parnzerror, "z0<0!",
		      *err, __LINE__, 0);

	 /* For z>z0 multiply n(z) with constant 0<c<1 */
	 if (z>self->par_nz[n_bin*self->Nnz_max+5]) {
            if (self->par_nz[n_bin*self->Nnz_max+6]<=0) res = 0;
            else res *= self->par_nz[n_bin*self->Nnz_max+6];
         }
         break;

      case hist :
	 testErrorRet(z<get_zmin(self, n_bin), redshift_zmax, "z smaller than zmin", *err, __LINE__, 0.0);
	 testErrorRet(z>get_zmax(self, n_bin), redshift_zmax, "z larger than zmax", *err, __LINE__, 0.0);

	 n = (self->Nnz[n_bin]-1)/2;

	 if (z>=get_zmin(self, n_bin) && z<self->par_nz[n_bin*self->Nnz_max+2]) {
	    /* z_0=z_min <= z < z_1: ---> n_0 */
	    res = self->par_nz[n_bin*self->Nnz_max+n+1];
	 } else {
	    for (m=2; m<n; m++) {
	       if (z>=self->par_nz[n_bin*self->Nnz_max+m] && z<self->par_nz[n_bin*self->Nnz_max+m+1]) break;
	    }
	    /* z_m <= z < z_{m+1} ---> n_m */
	    res = self->par_nz[n_bin*self->Nnz_max+n+m];
	 }
	 //fprintf(stderr, "%.3f %10.2f\n", z, res);
	 break;

      default :
	 *err = addErrorVA(redshift_unknown, "Unknown redshift distribution %d (type nofz_t). "
			   "See lensing.h for possible values", *err, __LINE__, self->nofz);
	 res = 0.0;
   }

   return res;
}
#undef __FUNC__

int change_prob(const redshift_t *avant, redshift_t *apres)
{
   return change_redshift(avant, apres);
}

int change_redshift(const redshift_t *avant, redshift_t *apres)
{
   int n, m;

   if (NCOEQ(avant, apres, Nzbin)) return 1;
   for (n=0; n<avant->Nzbin; n++) {
      if (NCOEQ(avant, apres, Nnz[n])) return 1;
      if (NCOEQ(avant, apres, nofz[n])) return 1;
      for (m=0; m<avant->Nnz[n]; m++) {
	 if (NCOCLOSE(avant, apres, par_nz[n*avant->Nnz_max+m])) return 1;
      }
   }
   return 0;
}

/* ============================================================ *
 * Redshift probability distribution of source galaxies.	*
 * Note that a single redshift sheet (beta_p=0) is		*
 * implemented in g_source.					*
 * ============================================================ */
#define __FUNC__ "prob"
double prob(redshift_t *self, double z, int n_bin, error **err)
{
   double res;
   redshiftANDint rANDi;

   if (z<get_zmin(self, n_bin) || z>get_zmax(self, n_bin)) return 0.0;

   testErrorRet(z<0, ce_negative, "Negative redshift z", *err, __LINE__ , 0);
		
   rANDi.self = self;
   rANDi.i    = n_bin;

   if (self->prob_norm==NULL) {
      self->prob_norm = calloc_err(self->Nzbin, sizeof(double), err);
      forwardError(*err, __LINE__, 0.0);
   }
   if (self->prob_norm[n_bin] <= 0.0) {
      self->prob_norm[n_bin] = 1.0/sm2_qromberg(prob_unnorm, (void*)&rANDi, get_zmin(self, n_bin),
						get_zmax(self, n_bin), 1.0e-6, err);
      forwardError(*err,__LINE__, 0);
   }

   res = self->prob_norm[n_bin]*prob_unnorm(z, (void*)&rANDi, err);
   forwardError(*err, __LINE__, 0);
   return res;
}
#undef __FUNC__

#define __FUNC__ "int_for_zmean"
double int_for_zmean(double z, void *intpar, error **err)
{
   double res;
   redshiftANDint *rANDi;
   int n_bin;

   rANDi = (redshiftANDint*)intpar;
   n_bin = rANDi->i;

   res = z*prob(rANDi->self, z, n_bin, err);
   forwardError(*err, __LINE__, 0);
   return res;	
}
#undef __FUNC__

#define __FUNC__ "zmean"
double zmean(redshift_t* self, int n_bin, error **err)
{
   double res;
   redshiftANDint rANDi;

   /* Single lens plane */
   if (self->nofz[n_bin]==single)
     return get_zmin(self, n_bin);   /* zmin = zmax = z0 */

   rANDi.self = self;
   rANDi.i    = n_bin;
   res = sm2_qromberg(int_for_zmean, (void*)&rANDi, get_zmin(self, n_bin), get_zmax(self, n_bin),
		      1.0e-6, err);
   forwardError(*err, __LINE__, 0);

   return res;	
}
#undef __FUNC__
