
#include "sn1a.h"
#include "cosmo.h"
#include "io.h"
#include "maths.h"
#include "errorlist.h"


#define CHI2_BIG 10000.0


void read_test(char *filename)
{
   cosmo_SN *cosmo;
   SnSample *TheSnSample;
   FILE *OUT;
   sm2_error *err = NULL;
   double c2;

   cosmo = set_cosmological_parameters_to_default_SN(&err);
   exitOnError(err, stderr);

   TheSnSample = SnSample_read(filename, &err);
   exitOnError(err, stderr);

   SetDl(cosmo, TheSnSample, &err);
   exitOnError(err, stderr);

   //out_SnSample(TheSnSample, stdout);

   OUT = fopen("dlum", "w");
   out_model(cosmo, OUT, &err);
   exitOnError(err, stderr);
   fclose(OUT);

   c2 = chi2_SN(cosmo, TheSnSample, NULL, 0, &err);
   exitOnError(err, stderr);
   printf("# c2 = %g\n", c2);

   free(cosmo);
   cosmo = set_cosmological_parameters_to_EdS_SN(&err);
   exitOnError(err, stderr);
   OUT = fopen("dlum_EdS", "w");
   out_model(cosmo, OUT, &err);
   exitOnError(err, stderr);
   fclose(OUT);
}

void acc_test(char *filename)
{
   cosmo_SN *previous, *current;
   SnSample *TheSnSample;
   FILE *OUT, *LOG;
   sm2_error *err = NULL;
   double c2, Om, OL;
   int error;

   TheSnSample = SnSample_read(filename, &err);
   exitOnError(err, stderr);

   previous = set_cosmological_parameters_to_default_SN(&err);
   exitOnError(err, stderr);

   OUT = fopen("chi2_OmOL", "w");
   LOG = fopen("log", "w");

   for (Om=0.01; Om<=1.0; Om+=0.01) {
      for (OL=0.01; OL<=1.7; OL+=0.01) {

	 error = 0;

	 /* set parameters */
	 current = copy_parameters_SN_only(previous, &err);
	 if (getErrorValue(err)!=noErr) {
	    printError(LOG, err);
	    err = NULL;
	    c2 = CHI2_BIG;
	    error = 1;
	    goto write_chi2;
	 }

	 current->cosmo->Omega_m  = Om;
	 current->cosmo->Omega_de = OL;
	 updateFrom_SN(previous, current, &err);
	 if (getErrorValue(err)!=noErr) {
	    printError(LOG, err);
	    err = NULL;
	    c2 = CHI2_BIG;
	    error = 1;
	    goto write_chi2;
	 }

	 SetDl(current, TheSnSample, &err);
	 if (getErrorValue(err)!=noErr) {
	    printError(LOG, err);
	    err = NULL;
	    c2 = CHI2_BIG;
	    error = 1;
	    goto write_chi2;
	 }


	 /* chi2 */
	 c2 = chi2_SN(current, TheSnSample, NULL, 0, &err);
	 if (getErrorValue(err)!=noErr) {
	    printError(LOG, err);
	    err = NULL;
	    c2 = CHI2_BIG;
	    error = 1;
	    goto write_chi2;
	 }


      write_chi2:
	 fprintf(OUT, "%f %f %10.4f 0.0\n", current->cosmo->Omega_m, current->cosmo->Omega_de, c2);


	 /* wrap up */
	 if (!error) {
	    free(previous);
	    previous = copy_parameters_SN_only(current, &err);
	    if (getErrorValue(err)!=noErr) { printError(LOG, err); continue; }
	    updateFrom_SN(previous, current, &err);
	    if (getErrorValue(err)!=noErr) { printError(LOG, err); continue; }
	 } else {
	    /* error occured: current might not be defined */
	    current = previous;
	 }

      }
   }

   fclose(OUT);
   fclose(LOG);
}

void print_dl()
{
   cosmo *self;
   error *myerr = NULL, **err;
   double z, dl, mu;

   err = &myerr;

   self = set_cosmological_parameters_to_default(err);
   exitOnError(*err, stderr);

   for (z=0; z<1.1; z+=0.01) {
      dl = D_lum(self, 1/(1.0+z), err);
      exitOnError(*err, stderr);
      mu = distance_module(self, dl, err);
      printf("%g %g %g\n", z, mu, dl);
      exitOnError(*err, stderr);
   }
}

void subtract_dl(char *filename)
{
   SnSample *sn;
   error *myerr = NULL, **err;
   cosmo_SN *self;
   int i;
   double sig0, sigF, sigbase;

   err = &myerr;

   self = set_cosmological_parameters_to_best_fit_SNLS_WMAP5(err);
   exitOnError(*err, stderr);

   sn = SnSample_read(filename, err);
   exitOnError(*err, stderr);
   SetDl(self, sn, err);
   exitOnError(*err, stderr);

   printf("# z   m-M   a(s-1)-bc  mu_c sig0 sigF\n");
   for (i=0; i<sn->Nsample; i++) {
      sigbase = dsqr(sn->int_disp) + dsqr(sn->sig_mu_pec_vel/sn->data[i].z);
      sig0 = sn->data[i].cov[0][0] + sigbase;
      sigF = DistModVariance(sn->data+i, self->Theta2) + sigbase;
      printf("%.4f  %g %g  %g  %g %g\n", sn->data[i].z, sn->data[i].musb + self->Theta2[0],
	     self->Theta2[1]*(sn->data[i].s-1.0) + self->Theta2[2]*sn->data[i].c,
	     sn->data[i].mu_c, sqrt(sig0), sqrt(sigF));
   }

   free_parameters_SN(&self);

}

void subtract_dl2(char *filename)
{
   SnSample *sn;
   error *myerr = NULL, **err;
   cosmo_SN *self;
   int i;
   double sigbase, sigF;

   err = &myerr;

   self = set_cosmological_parameters_to_best_fit_Union(err);
   exitOnError(*err, stderr);

   sn = SnSample_read(filename, err);
   exitOnError(*err, stderr);
   SetDl(self, sn, err);
   exitOnError(*err, stderr);

   printf("# z   mu delta(mu) 5log(dL/10pc)\n");
   for (i=0; i<sn->Nsample; i++) {
      sigbase = dsqr(sn->int_disp) + dsqr(sn->sig_mu_pec_vel/sn->data[i].z);
      sigF = DistModVariance(sn->data+i, self->Theta2) + sigbase;
      printf("%.4f  %g %g  %g\n", sn->data[i].z, sn->data[i].musb + self->Theta2[0] +
	     self->Theta2[1]*(sn->data[i].s-1.0) + self->Theta2[2]*sn->data[i].c,
	     sqrt(sigF), sn->data[i].mu_c);
   }

   free_parameters_SN(&self);

}


int main(int argc, char *argv[])
{

   //print_dl();

   subtract_dl2(argv[1]);

   //read_test(argv[1]);

   //acc_test(argv[1]);

   return 0;
}
