/* ============================================================ *
 * decomp_eb.c							*
 * Martin Kilbinger, Liping Fu 2008, 2009			*
 * ============================================================ */

#include "decomp_eb.h"

/* First-kind Chebyshev polynomial of order n */
#define EPS 1.0e-10
#define __FUNC__ "Cheby"
double Cheby(double x, int n, error **err)
{
   double Cn;

   //printf("Cheby: x = %.20f\n", x);

   testErrorRetVA(x<-1-EPS || x>1+EPS, mr_range, "x = %g out of range", *err, __LINE__, -1, x);

   if (x<-1) x = -1.0;
   if (x>1) x = 1.0;

   Cn = cos (n * acos (x));
   return Cn;
}
#undef EPS
#undef __FUNC__

/* Second-kind Chebyshev polynomial of order n */
#define EPS 1.0e-10
#define __FUNC__ "Cheby2"
double Cheby2(double x, int n, error **err)
{
   double Un;

   testErrorRetVA(x<-1-EPS || x>1+EPS, mr_range, "x = %g out of range", *err, __LINE__, -1, x);

   if (x<-1) x = -1.0;
   if (x>1) x = 1.0;

   if (x == 1) {
      Un = n + 1;
   } else if (x == -1) {
      Un = pow(-1.0, n) * (n + 1.0);
   } else {
      Un = sin((n+1.0)*acos(x))/sin(acos (x));
   }

   return Un;
}
#undef EPS
#undef __FUNC__

/* Legendre polynomial of order n */
double Legen(double x, int n)
{

   switch (n) {
      case 0:
	 return 1.0;
      case 1:
	 return x;
      default:
	 return (2.0*n-1.0)/(double)n*x*Legen(x,n-1)-(n-1.0)/(double)n*Legen(x,n-2);  
   }
}

/* General basis function of order n */
#define EPS 1.0e-10
#define __FUNC__ "C"
double C(double x, int n, poly_t poly, error **err)
{
   double c;

   testErrorRetVA(x<-1-EPS || x>1+EPS, mr_range, "x = %g out of range", *err, __LINE__, -1, x);

   if (x<-1) x = -1.0;
   if (x>1) x = 1.0;

   switch (poly) {
      case cheby :
	 c = Cheby(x, n, err);
	 forwardError(*err, __LINE__, -1.0);
	 break;
      case cheby2 :
	 c = Cheby2(x, n, err);
	 forwardError(*err, __LINE__, -1.0);
	 break;
      case legen :
	 c = Legen(x, n);
	 break;
      default :
	 *err = addErrorVA(mr_poly, "Unknown polynomial type %d", *err, __LINE__, poly);
	 return -1.0;
   }

   return c;
}
#undef __FUNC__
#undef EPS


/* ============================================================ *
 * The filter function for the generalised ring statistics.	*
 * FK09 (11).							*
 * ============================================================ */
#define	 EPS 1.0e-6
#define __FUNC__ "Tp"
double Tp(double x, const double *a, int N, poly_t poly, error **err)
{
   int n;
   double res, Cn, Cnm1=0.0, Cnm2; 

   if (x<-1-EPS || x>+1+EPS) return 0;

   /* NEW! */
   //if (x<-1) x = -1;
   //if (x>+1) x = +1;


   testErrorRetVA(N<=0, mr_range, "N has to be larger than zero but is %d", *err, __LINE__, 0.0, N);

   testErrorRetVA(x>+1+EPS, mr_range, "x=%.10f out of range", *err, __LINE__, 0.0, x);
   testErrorRetVA(x<-1-EPS, mr_range, "x=%.10f out of range", *err, __LINE__, 0.0, x);

   if (poly==cheby || poly==cheby2) {

      Cnm2 = 1.0;
      if (poly==cheby)       Cnm1 = x;
      else if (poly==cheby2) Cnm1 = 2.0*x;

      res  = 0.0;
      res += a[0] * Cnm2;
      if (N==1) return res;
      res += a[1] * Cnm1;
      if (N==2) return res;
	
      for (n=2; n<N; n++) {
	 /* Chebyshev (both T,U) recurrence relation *
	  * C_n(x) = 2 x C_{n-1}(x) - C_{n-2}(x)     */
	 Cn   = 2.0*x*Cnm1 - Cnm2;
	 res += a[n] * Cn;
	 Cnm2 = Cnm1;
	 Cnm1 = Cn;
      }

   } else {

      for (n=0,res=0.0; n<N; n++) {
	 Cn   = C(x, n, poly, err);
	 forwardError(*err, __LINE__, -1);
	 res += a[n] * Cn;
      }

   }

   return res;
}
#undef __FUNC__
#undef EPS

/* ============================================================ *
 * FK09 (24).							*
 * ============================================================ */
#define __FUNC__ "Fn0"
double Fn0(double x, int n, poly_t poly, error **err)
{
   double fn0, denom=0.0;
   int nn;


   if (n<0) nn = -n;
   else nn = n;

   if (poly==cheby) {
      if (nn==0) return x + 1.0;
      if (nn==1) return 0.5*(x*x - 1.0);
   } else if (poly==cheby2) {
      if (n==-1) return 0.0;  /* n, not nn: no symmetry for cheby2! */
   } else {
      *err = addErrorVA(mr_poly, "Tm for polynomial type %d not defined", *err, __LINE__, poly);
      return 0.0;
   }

   fn0 = 0.0;

   if (nn%2==0) fn0 += 1.0;   /* n even */
   else fn0 -= 1.0;           /* n odd  */

   //printf("# ** 1 %g\n", fn0);

   fn0 += x*Cheby(x, nn, err);
   forwardError(*err, __LINE__, 0.0);

   //printf("# ** 2 %g\n", fn0);

   if (poly==cheby) fn0 += n*(1 - x*x)*Cheby2(x, n-1, err);
   else if (poly==cheby2) fn0 -= (1 - x*x)*Cheby2(x, n-1, err);
   forwardError(*err, __LINE__, 0.0);

   //printf("# ** 3 %g\n", fn0);

   if (poly==cheby) denom = 1.0 - (double)(n*n); 
   else if (poly==cheby2) denom = 1.0+(double)n;         /* n, not nn! */

   fn0 /= denom;

   return fn0;
}
#undef __FUNC__

/* ============================================================ *
 * FK09 (26)							*
 * ============================================================ */
#define __FUNC__ "Fnnu"
void Fnnu(double x, int n, poly_t poly, double Fn[], error **err)
{
   double f0, fp1, fp2, fp3, fm1, fm2, fm3;

   f0  = Fn0(x, n, poly, err);      forwardError(*err, __LINE__,);
   fp1 = Fn0(x, n+1, poly, err);    forwardError(*err, __LINE__,);
   fp2 = Fn0(x, n+2, poly, err);    forwardError(*err, __LINE__,);
   fp3 = Fn0(x, n+3, poly, err);    forwardError(*err, __LINE__,);
   fm1 = Fn0(x, n-1, poly, err);    forwardError(*err, __LINE__,);
   fm2 = Fn0(x, n-2, poly, err);    forwardError(*err, __LINE__,);
   fm3 = Fn0(x, n-3, poly, err);    forwardError(*err, __LINE__,);

   Fn[0] = f0;
   Fn[1] = 0.5*(fp1 + fm1);
   Fn[2] = 0.25*(fp2 + 2.0*f0 + fm2);
   Fn[3] = 0.125*(fp3 + 3.0*fp1 + 3.0*fm1 + fm3);
}
#undef __FUNC__

/* ============================================================ *
 * FK09 (23)							*
 * ============================================================ */
#define __FUNC__ "alpha_explicit"
#define EPS 1.0e-10
double alpha_explicit(double x, int n, int N, double R, poly_t poly, error **err)
{
   double r, Fn[4], alpha;

   testErrorRet(poly!=cheby && poly!=cheby2, mr_poly,
		"Explicit expression only obtained so far for cheby1 and cheby2",
		*err, __LINE__, 0.0);

   if (x<-1.0+EPS) return 0.0;
   assert(x>-1.0);

   r = 3.0/dsqr(x + R);
   Fnnu(x, n, poly, Fn, err);
   forwardError(*err, __LINE__, 0.0);

   alpha  = R*(1.0-r*R*R)*Fn[0];
   alpha += (1.0 - 3.0*r*R*R)*Fn[1];
   alpha += -3.0*r*R*Fn[2];
   alpha += -r*Fn[3];
   alpha *= 4.0/3.0*r;

   return alpha;
}
#undef __FUNC__

/* ============================================================ *
 * The filter function T_- obtained from T_+, FK09 (20).	*
 * ============================================================ */
#define EPS 1.0e-10
#define __FUNC__ "Tm"
double Tm(double x, const double *a, int N, poly_t poly, double R, error **err)
{
   double tm, c, alph;
   int n;

   testErrorRetVA(x<-1-EPS, mr_range, "x=%.20f smaller than -1", *err, __LINE__, 0.0, x);

   for (n=0,tm=0.0; n<N; n++) {
      c    = C(x, n, poly, err);                forwardError(*err, __LINE__, 0.0);
      alph = alpha_explicit(x, n, N, R, poly, err);
      forwardError(*err, __LINE__, 0.0);
      
      tm  += a[n]*(c + alph);
   }
   return tm;
}
#undef __FUNC__

/* Returns the generalised ring statistic using a as the coefficients *
 * of the T_+ decomposition, see FK09 (4, 8, 11)		      *
 * If a==NULL, Z+ from SK07 is used.				      */
#define __FUNC__ "RR"
double RR(cosmo_lens *lens, double THETA_MIN, double THETA_MAX, const double *a, int N,
	  poly_t poly, mring_combinef_t combinef, int pm, error **err)
{
   double theta, dtheta, res[2]={0,0}, A, B, x, summand;

   testErrorRetVA(abs(pm)!=1, mr_incompatible, "pm=%d not valid, has to be +1 or -1",
		  *err, __LINE__, 0.0, pm);

   if (a==NULL) {
#ifdef __MRING_H
      R_from_xi(lens, THETA_MIN/THETA_MAX, THETA_MAX, res, res+1, err);
      forwardError(*err, __LINE__, 0.0);
      return res[0];
#else
      *err = addError(mr_null, "Coefficients a=NULL", *err, __LINE__);
      return 0.0;
#endif
   }


   /* Decrease step size dtheta for more precision,     *
    * in particular, to make (numerical) B-mode smaller	*/
   dtheta = 1.0*arcsec/10.0;


   /* FK09 (8) */
   A = (THETA_MAX - THETA_MIN)/2.0;
   B = (THETA_MAX + THETA_MIN)/2.0;

   for (theta=THETA_MIN,res[0]=0.0; theta<=THETA_MAX; theta+=dtheta) {

      x        = (theta-B)/A;

      summand  = xi(lens, pm, theta, 0, 0, err)*theta/dsqr(THETA_MAX);
      forwardError(*err, __LINE__, -1.0);

      if (pm==+1)      summand *= Tp(x, a, N, poly, err);
      else if (pm==-1) summand *= Tm(x, a, N, poly, B/A, err);
      forwardError(*err, __LINE__, -1.0);

      res[0] += summand;
   }

   testErrorRet(!finite(res[0]), math_infnan, "R_E is not finite", *err, __LINE__,
		  0.0);

   return res[0]*dtheta;
}
#undef __FUNC__
